from __future__ import annotations

from pymbolic.mapper.evaluator import evaluate_kw
from pymbolic.mapper.flattener import FlattenMapper
from pymbolic.mapper.stringifier import StringifyMapper
from pymbolic.typing import Expression


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging
from functools import reduce

import pytest
from testlib import generate_random_expression

from pytools.lex import ParseError

import pymbolic.primitives as prim
from pymbolic import parse
from pymbolic.mapper import IdentityMapper, WalkMapper
from pymbolic.mapper.dependency import CachedDependencyMapper, DependencyMapper


logger = logging.getLogger(__name__)


# {{{ utilities

def assert_parsed_same_as_python(expr_str):
    # makes sure that has only one line
    expr_str, = expr_str.split("\n")

    import ast

    from pymbolic.interop.ast import ASTToPymbolic
    ast2p = ASTToPymbolic()

    try:
        expr_parsed_by_python = ast2p(ast.parse(expr_str).body[0].value)
    except SyntaxError:
        with pytest.raises(ParseError):
            parse(expr_str)
    else:
        expr_parsed_by_pymbolic = parse(expr_str)
        assert expr_parsed_by_python == expr_parsed_by_pymbolic


def assert_parse_roundtrip(expr_str):
    from pymbolic.mapper.stringifier import StringifyMapper
    expr = parse(expr_str)
    strified = StringifyMapper()(expr)

    assert strified == expr_str, (strified, expr_str)

# }}}


EXPRESSION_COLLECTION = [
    parse("(x[2]+y.data)*(x+z)**3"),
    parse("(~x)//2 | (y >> 2) & (z << 3)"),
    parse("x and (not y or z)"),
    parse("x if not (y and z) else x+1"),
]


# {{{ test_integer_power

def test_integer_power():
    from pymbolic.algorithm import integer_power

    for base, expn in [
            (17, 5),
            (17, 2**10),
            (13, 20),
            (13, 1343),
            ]:
        assert base**expn == integer_power(base, expn)

# }}}


# {{{ test_expand

def test_expand():
    from pymbolic import expand, var

    x = var("x")
    u = (x+1)**5
    expand(u)

# }}}


# {{{ test_substitute

def test_substitute():
    from pymbolic import evaluate, parse, substitute
    u = parse("5+x.min**2")
    xmin = parse("x.min")
    assert evaluate(substitute(u, {xmin: 25})) == 630

# }}}


# {{{ test_no_comparison

def test_no_comparison():
    from pymbolic import parse

    x = parse("17+3*x")
    y = parse("12-5*y")

    def expect_typeerror(f):
        try:
            f()
        except TypeError:
            pass
        else:
            raise AssertionError

    expect_typeerror(lambda: x < y)
    expect_typeerror(lambda: x <= y)
    expect_typeerror(lambda: x > y)
    expect_typeerror(lambda: x >= y)

# }}}


# {{{ test_structure_preservation

def test_structure_preservation():
    x = prim.Sum((5, 7))
    x2 = IdentityMapper()(x)
    assert x == x2

# }}}


# {{{ test_sympy_interaction

def test_sympy_interaction():
    pytest.importorskip("sympy")

    import sympy as sp

    x, y = sp.symbols("x y")
    f = sp.Function("f")

    s1_expr = 1/f(x/sp.sqrt(x**2+y**2)).diff(x, 5)  # pylint:disable=not-callable

    from pymbolic.interop.sympy import PymbolicToSympyMapper, SympyToPymbolicMapper
    s2p = SympyToPymbolicMapper()
    p2s = PymbolicToSympyMapper()

    p1_expr = s2p(s1_expr)
    s2_expr = p2s(p1_expr)

    assert sp.ratsimp(s1_expr - s2_expr) == 0

    p2_expr = s2p(s2_expr)
    s3_expr = p2s(p2_expr)

    assert sp.ratsimp(s1_expr - s3_expr) == 0

# }}}


# {{{ fft

def test_fft_with_floats():
    numpy = pytest.importorskip("numpy")
    import numpy.linalg as la

    from pymbolic.algorithm import fft, ifft

    for n in [2**i for i in range(4, 10)]+[17, 12, 948]:
        a = numpy.random.rand(n) + 1j*numpy.random.rand(n)
        f_a = fft(a)
        a2 = ifft(f_a)
        assert la.norm(a-a2) < 1e-10

        f_a_numpy = numpy.fft.fft(a)
        assert la.norm(f_a-f_a_numpy) < 1e-10


class NearZeroKiller(IdentityMapper):
    def map_constant(self, expr):
        if isinstance(expr, complex):
            r = expr.real
            i = expr.imag
            if abs(r) < 1e-15:
                r = 0
            if abs(i) < 1e-15:
                i = 0
            return complex(r, i)
        else:
            return expr


def test_fft():
    numpy = pytest.importorskip("numpy")

    from pymbolic import var
    from pymbolic.algorithm import fft, sym_fft

    vars = numpy.array([var(chr(97+i)) for i in range(16)], dtype=object)
    logger.info("vars: %s", vars)

    logger.info("fft: %s", fft(vars))
    traced_fft = sym_fft(vars)

    from pymbolic.mapper.c_code import CCodeMapper
    from pymbolic.mapper.stringifier import PREC_NONE
    ccm = CCodeMapper()

    code = [ccm(tfi, PREC_NONE) for tfi in traced_fft]

    for cse_name, cse_str in enumerate(ccm.cse_name_list):
        logger.info("%s = %s", cse_name, cse_str)

    for i, line in enumerate(code):
        logger.info("result[%d] = %s", i, line)

# }}}


# {{{ test_sparse_multiply

def test_sparse_multiply():
    numpy = pytest.importorskip("numpy")
    pytest.importorskip("scipy")
    import scipy.sparse as ss

    la = numpy.linalg

    mat = numpy.random.randn(10, 10)
    s_mat = ss.csr_matrix(mat)

    vec = numpy.random.randn(10)
    mat_vec = s_mat*vec

    from pymbolic.algorithm import csr_matrix_multiply
    mat_vec_2 = csr_matrix_multiply(s_mat, vec)

    assert la.norm(mat_vec-mat_vec_2) < 1e-14

# }}}


# {{{ parser

def test_parser():
    from pymbolic import parse
    parse("(2*a[1]*b[1]+2*a[0]*b[0])*(hankel_1(-1,sqrt(a[1]**2+a[0]**2)*k) "
            "-hankel_1(1,sqrt(a[1]**2+a[0]**2)*k))*k /(4*sqrt(a[1]**2+a[0]**2)) "
            "+hankel_1(0,sqrt(a[1]**2+a[0]**2)*k)")
    logger.info("%r", parse("d4knl0"))
    logger.info("%r", parse("0."))
    logger.info("%r", parse("0.e1"))
    assert parse("0.e1") == 0
    assert parse("1e-12") == 1e-12
    logger.info("%r", parse("a >= 1"))
    logger.info("%r", parse("a <= 1"))

    logger.info("%r", parse(":"))
    logger.info("%r", parse("1:"))
    logger.info("%r", parse(":2"))
    logger.info("%r", parse("1:2"))
    logger.info("%r", parse("::"))
    logger.info("%r", parse("1::"))
    logger.info("%r", parse(":1:"))
    logger.info("%r", parse("::1"))
    logger.info("%r", parse("3::1"))
    logger.info("%r", parse(":5:1"))
    logger.info("%r", parse("3:5:1"))

    assert_parse_roundtrip("()")
    assert_parse_roundtrip("(3,)")

    assert_parse_roundtrip("[x + 3, 3, 5]")
    assert_parse_roundtrip("[]")
    assert_parse_roundtrip("[x]")

    assert_parse_roundtrip("g[i, k] + 2.0*h[i, k]")
    parse("g[i,k]+(+2.0)*h[i, k]")

    logger.info("%r", parse("a - b - c"))
    logger.info("%r", parse("-a - -b - -c"))
    logger.info("%r", parse("- - - a - - - - b - - - - - c"))

    logger.info("%r", parse("~(a ^ b)"))
    logger.info("%r", parse("(a | b) | ~(~a & ~b)"))

    logger.info("%r", parse("3 << 1"))
    logger.info("%r", parse("1 >> 3"))

    logger.info(parse("3::1"))

    assert parse("e1") == prim.Variable("e1")
    assert parse("d1") == prim.Variable("d1")

    from pymbolic import variables
    f, x, y, z = variables("f x y z")
    assert parse("f((x,y),z)") == f((x, y), z)
    assert parse("f((x,),z)") == f((x,), z)
    assert parse("f(x,(y,z),z)") == f(x, (y, z), z)

    assert parse("f(x,(y,z),z, name=15)") == f(x, (y, z), z, name=15)
    assert parse("f(x,(y,z),z, name=15, name2=17)") == f(
            x, (y, z), z, name=15, name2=17)

    assert_parsed_same_as_python("5+i if i>=0 else (0 if i<-1 else 10)")
    assert_parsed_same_as_python("0 if 1 if 2 else 3 else 4")
    assert_parsed_same_as_python("0 if (1 if 2 else 3) else 4")
    assert_parsed_same_as_python("(2, 3,)")
    assert_parsed_same_as_python("-3**0.5")
    assert_parsed_same_as_python("1/2/7")

    with pytest.deprecated_call():
        parse("1+if(0, 1, 2)")

    assert eval(str(parse("1729 if True or False else 42"))) == 1729

# }}}


# {{{ test_mappers

def test_mappers():
    from pymbolic import variables
    f, x, y, z = variables("f x y z")

    for expr in [
            f(x, (y, z), name=z**2)
            ]:
        str(expr)
        IdentityMapper()(expr)
        WalkMapper()(expr)
        DependencyMapper()(expr)


# }}}


# {{{ test_func_dep_consistency

def test_func_dep_consistency():
    from pymbolic import var
    f = var("f")
    x = var("x")
    dep_map = DependencyMapper(include_calls="descend_args")
    assert dep_map(f(x)) == {x}
    assert dep_map(f(x=x)) == {x}

# }}}


# {{{ test_conditions

def test_conditions():
    from pymbolic import var
    x = var("x")
    y = var("y")
    assert str(x.eq(y).and_(x.le(5))) == "x == y and x <= 5"

# }}}


# {{{ test_graphviz

def test_graphviz():
    from pymbolic import parse
    expr = parse("(2*a[1]*b[1]+2*a[0]*b[0])*(hankel_1(-1,sqrt(a[1]**2+a[0]**2)*k) "
            "-hankel_1(1,sqrt(a[1]**2+a[0]**2)*k))*k /(4*sqrt(a[1]**2+a[0]**2)) "
            "+hankel_1(0,sqrt(a[1]**2+a[0]**2)*k)")

    from pymbolic.mapper.graphviz import GraphvizMapper
    gvm = GraphvizMapper()
    gvm(expr)
    logger.info("%s", gvm.get_dot_code())

# }}}


# {{{ geometric algebra

@pytest.mark.parametrize("dims", [2, 3, 4, 5])
# START_GA_TEST
def test_geometric_algebra(dims):
    pytest.importorskip("numpy")

    import numpy as np

    from pymbolic.geometric_algebra import MultiVector as MV  # noqa

    vec1 = MV(np.random.randn(dims))
    vec2 = MV(np.random.randn(dims))
    vec3 = MV(np.random.randn(dims))
    vec4 = MV(np.random.randn(dims))
    vec5 = MV(np.random.randn(dims))

    # Fundamental identity
    assert ((vec1 ^ vec2) + (vec1 | vec2)).close_to(vec1*vec2)

    # Antisymmetry
    assert (vec1 ^ vec2 ^ vec3).close_to(- vec2 ^ vec1 ^ vec3)

    vecs = [vec1, vec2, vec3, vec4, vec5]

    if len(vecs) > dims:
        from operator import xor as outer
        assert reduce(outer, vecs).close_to(0)

    assert (vec1.inv()*vec1).close_to(1)
    assert (vec1*vec1.inv()).close_to(1)
    assert ((1/vec1)*vec1).close_to(1)
    assert (vec1/vec1).close_to(1)

    for a, b, c in [
            (vec1, vec2, vec3),
            (vec1*vec2, vec3, vec4),
            (vec1, vec2*vec3, vec4),
            (vec1, vec2, vec3*vec4),
            (vec1, vec2, vec3*vec4*vec5),
            (vec1, vec2*vec1, vec3*vec4*vec5),
            ]:

        # Associativity
        assert ((a*b)*c).close_to(a*(b*c))
        assert ((a ^ b) ^ c).close_to(a ^ (b ^ c))
        # The inner product is not associative.

        # scalar product
        assert ((c*b).project(0)) .close_to(b.scalar_product(c))
        assert ((c.rev()*b).project(0)) .close_to(b.rev().scalar_product(c))
        assert ((b.rev()*b).project(0)) .close_to(b.norm_squared())

        assert b.norm_squared() >= 0
        assert c.norm_squared() >= 0

        # Cauchy's inequality
        assert b.scalar_product(c) <= abs(b)*abs(c) + 1e-13

        # contractions

        # (3.18) in [DFM]
        assert abs(b.scalar_product(a ^ c) - (b >> a).scalar_product(c)) < 1e-13

        # duality, (3.20) in [DFM]
        assert ((a ^ b) << c) .close_to(a << (b << c))

        # two definitions of the dual agree: (1.2.26) in [HS]
        # and (sec 3.5.3) in [DFW]
        assert (c << c.I.rev()).close_to(c | c.I.rev())

        # inverse
        for div in [*b.gen_blades(), vec1, vec1.I]:
            assert (div.inv()*div).close_to(1)
            assert (div*div.inv()).close_to(1)
            assert ((1/div)*div).close_to(1)
            assert (div/div).close_to(1)
            assert ((c/div)*div).close_to(c)
            assert ((c*div)/div).close_to(c)

        # reverse properties (Sec 2.9.5 [DFM])
        assert c.rev().rev() == c
        assert (b ^ c).rev() .close_to(c.rev() ^ b.rev())

        # dual properties
        # (1.2.26) in [HS]
        assert c.dual() .close_to(c | c.I.rev())
        assert c.dual() .close_to(c*c.I.rev())

        # involution properties (Sec 2.9.5 DFW)
        assert c.invol().invol() == c
        assert (b ^ c).invol() .close_to(b.invol() ^ c.invol())

        # commutator properties

        # Jacobi identity (1.1.56c) in [HS] or (8.2) in [DFW]
        assert (a.x(b.x(c)) + b.x(c.x(a)) + c.x(a.x(b))).close_to(0)

        # (1.57) in [HS]
        assert a.x(b*c) .close_to(a.x(b)*c + b*a.x(c))
# END_GA_TEST

# }}}


# {{{ test_ast_interop

def test_ast_interop():
    src = """
    def f():
        xx = 3*y + z * (12 if x < 13 else 13)
        yy = f(x, y=y)
    """

    import ast
    mod = ast.parse(src.replace("\n    ", "\n"))

    logger.info("%s", ast.dump(mod))

    from pymbolic.interop.ast import ASTToPymbolic
    ast2p = ASTToPymbolic()

    for f in mod.body:
        if not isinstance(f, ast.FunctionDef):
            continue

        for stmt in f.body:
            if not isinstance(stmt, ast.Assign):
                continue

            lhs, = stmt.targets
            lhs = ast2p(lhs)
            rhs = ast2p(stmt.value)

            logger.info("lhs %s rhs %s", lhs, rhs)

# }}}


# {{{ test_compile

def test_compile():
    from pymbolic import compile, parse
    code = compile(parse("x ** y"), ["x", "y"])
    assert code(2, 5) == 32

    # Test pickling of compiled code.
    import pickle
    code = pickle.loads(pickle.dumps(code))
    assert code(3, 3) == 27

# }}}


# {{{ test_pickle

def test_pickle():
    from pickle import dumps, loads
    for expr in EXPRESSION_COLLECTION:
        pickled = loads(dumps(expr))
        assert hash(expr) == hash(pickled)
        assert expr == pickled


class OldTimeyExpression(prim.ExpressionNode):
    init_arg_names = ()

    def __getinitargs__(self):
        return ()


def test_pickle_backward_compat():
    from pickle import dumps, loads

    expr = 3*OldTimeyExpression()
    pickled = loads(dumps(expr))
    with pytest.warns(DeprecationWarning):
        assert hash(expr) == hash(pickled)
    with pytest.warns(DeprecationWarning):
        assert expr == pickled
# }}}


# {{{ test_unifier

def test_unifier():
    from pymbolic import var
    from pymbolic.mapper.unifier import UnidirectionalUnifier
    a, b, c, d, e, f = (var(s) for s in "abcdef")

    def match_found(records, eqns):
        for record in records:
            if eqns <= set(record.equations):
                return True
        return False

    recs = UnidirectionalUnifier("abc")(a+b*c, d+e*f)
    assert len(recs) == 2
    assert match_found(recs, {(a, d), (b, e), (c, f)})
    assert match_found(recs, {(a, d), (b, f), (c, e)})

    recs = UnidirectionalUnifier("abc")(a+b, d+e+f)
    assert len(recs) == 6
    assert match_found(recs, {(a, d), (b, e+f)})
    assert match_found(recs, {(a, e), (b, d+f)})
    assert match_found(recs, {(a, f), (b, d+e)})
    assert match_found(recs, {(b, d), (a, e+f)})
    assert match_found(recs, {(b, e), (a, d+f)})
    assert match_found(recs, {(b, f), (a, d+e)})

    vals = [var("v" + str(i)) for i in range(100)]
    recs = UnidirectionalUnifier("a")(sum(vals[1:]) + a, sum(vals))
    assert len(recs) == 1
    assert match_found(recs, {(a, var("v0"))})

    recs = UnidirectionalUnifier("abc")(a+b+c, d+e)
    assert len(recs) == 0

    recs = UnidirectionalUnifier("abc")(f(a+b, f(a+c)), f(b+c, f(b+d)))
    assert len(recs) == 1
    assert match_found(recs, {(a, b), (b, c), (c, d)})

# }}}


# {{{ test_long_sympy_mapping

def test_long_sympy_mapping():
    sp = pytest.importorskip("sympy")
    from pymbolic.interop.sympy import SympyToPymbolicMapper
    SympyToPymbolicMapper()(sp.sympify(int(10**20)))
    SympyToPymbolicMapper()(sp.sympify(10))

# }}}


# {{{ test_stringifier_preserve_shift_order

def test_stringifier_preserve_shift_order():
    for expr in [
            parse("(a << b) >> 2"),
            parse("a << (b >> 2)")
            ]:
        assert parse(str(expr)) == expr

# }}}


# {{{ test_latex_mapper

LATEX_TEMPLATE = r"""\documentclass{article}
\usepackage{amsmath}

\begin{document}
%s
\end{document}"""


def test_latex_mapper():
    from pymbolic import parse
    from pymbolic.mapper.stringifier import LaTeXMapper, StringifyMapper

    tm = LaTeXMapper()
    sm = StringifyMapper()

    equations = []

    def add(expr):
        # Add an equation to the list of tests.
        equations.append(r"\[{}\] % from: {}".format(tm(expr), sm(expr)))

    add(parse("a * b + c"))
    add(parse("f(a,b,c)"))
    add(parse("a ** b ** c"))
    add(parse("(a | b) ^ ~c"))
    add(parse("a << b"))
    add(parse("a >> b"))
    add(parse("a[i,j,k]"))
    add(parse("a[1:3]"))
    add(parse("a // b"))
    add(parse("not (a or b) and c"))
    add(parse("(a % b) % c"))
    add(parse("(a >= b) or (b <= c)"))
    add(prim.Min((1,)) + prim.Max((1, 2)))
    add(prim.Substitution(prim.Variable("x") ** 2, ("x",), (2,)))
    add(prim.Derivative(parse("x**2"), ("x",)))

    # Run LaTeX and ensure the file compiles.
    import os
    import shutil
    import subprocess
    import tempfile

    latex_dir = tempfile.mkdtemp("pymbolic")

    try:
        tex_file_path = os.path.join(latex_dir, "input.tex")

        with open(tex_file_path, "w") as tex_file:
            contents = LATEX_TEMPLATE % "\n".join(equations)
            tex_file.write(contents)

        try:
            subprocess.check_output(
                    ["latex",
                     "-interaction=nonstopmode",
                     "-output-directory=%s" % latex_dir,
                     tex_file_path],
                    universal_newlines=True)
        except FileNotFoundError:
            pytest.skip("latex command not found")
        except subprocess.CalledProcessError as err:
            raise AssertionError(str(err.output)) from None

    finally:
        shutil.rmtree(latex_dir)

# }}}


# {{{ test_flop_counter

def test_flop_counter():
    x = prim.Variable("x")
    y = prim.Variable("y")
    z = prim.Variable("z")

    subexpr = prim.make_common_subexpression(3 * (x**2 + y + z))
    expr = 3*subexpr + subexpr

    from pymbolic.mapper.flop_counter import CSEAwareFlopCounter, FlopCounter
    assert FlopCounter()(expr) == 4 * 2 + 2

    assert CSEAwareFlopCounter()(expr) == 4 + 2

# }}}


# {{{ test_make_sym_vector

def test_make_sym_vector():
    numpy = pytest.importorskip("numpy")
    from pymbolic.primitives import make_sym_vector

    assert len(make_sym_vector("vec", 2)) == 2
    assert len(make_sym_vector("vec", numpy.int32(2))) == 2
    assert len(make_sym_vector("vec", [1, 2, 3])) == 3

# }}}


# {{{ test_multiplicative_stringify_preserves_association

def test_multiplicative_stringify_preserves_association():
    for inner in ["*", " / ", " // ", " % "]:
        for outer in ["*", " / ", " // ", " % "]:
            if outer == inner:
                continue

            assert_parse_roundtrip(f"x{outer}(y{inner}z)")
            assert_parse_roundtrip(f"(y{inner}z){outer}x")

    assert_parse_roundtrip("(-1)*(((-1)*x) / 5)")

# }}}


# {{{ test_differentiator_flags_for_nonsmooth_and_discontinuous

def test_differentiator_flags_for_nonsmooth_and_discontinuous():
    import pymbolic.functions as pf
    from pymbolic.mapper.differentiator import differentiate

    x = prim.Variable("x")

    with pytest.raises(ValueError):
        differentiate(pf.fabs(x), x)

    result = differentiate(pf.fabs(x), x, allowed_nonsmoothness="continuous")
    assert result == pf.sign(x)

    with pytest.raises(ValueError):
        differentiate(pf.sign(x), x)

    result = differentiate(pf.sign(x), x, allowed_nonsmoothness="discontinuous")
    assert result == 0

# }}}


# {{{ test_diff_cse

def test_diff_cse():
    from pymbolic import evaluate_kw
    from pymbolic.mapper.differentiator import differentiate

    m = prim.Variable("math")

    x = prim.Variable("x")
    cse = prim.make_common_subexpression(x**2 + 1)
    expr = m.attr("exp")(cse)*m.attr("sin")(cse**2)

    diff_result = differentiate(expr, x)

    import math
    from functools import partial
    my_eval = partial(evaluate_kw, math=math)

    x0 = 5
    h = 0.001
    fprime = my_eval(diff_result, x=x0)
    fprime_num_1 = (my_eval(expr, x=x0+h) - my_eval(expr, x=x0-h))/(2*h)
    fprime_num_2 = (my_eval(expr, x=x0+0.5*h) - my_eval(expr, x=x0-0.5*h))/h

    err1 = abs(fprime - fprime_num_1)/abs(fprime)
    err2 = abs(fprime - fprime_num_2)/abs(fprime)

    assert err2 < 1.1 * 0.5**2 * err1

# }}}


# {{{ test_coefficient_collector

def test_coefficient_collector():
    from pymbolic.mapper.coefficient import CoefficientCollector
    x = prim.Variable("x")
    y = prim.Variable("y")
    z = prim.Variable("z")

    cc = CoefficientCollector([x.name, y.name])
    assert cc(2*x + y) == {x: 2, y: 1}
    assert cc(2*x + y - z) == {x: 2, y: 1, 1: -z}
    assert cc(x/2 + z**2) == {x: prim.Quotient(1, 2), 1: z**2}

# }}}


# {{{ test_np_bool_handling

def test_np_bool_handling():
    from pymbolic.mapper.evaluator import evaluate
    numpy = pytest.importorskip("numpy")
    expr = prim.LogicalNot(numpy.bool_(False))
    assert evaluate(expr) is True

# }}}


# {{{ test_mapper_method_of_parent_class

def test_mapper_method_of_parent_class():
    class SpatialConstant(prim.Variable):
        mapper_method = "map_spatial_constant"

    class MyMapper(IdentityMapper):
        def map_spatial_constant(self, expr):
            return 2*expr

    c = SpatialConstant("k")

    assert MyMapper()(c) == 2*c
    assert IdentityMapper()(c) == c

# }}}


# {{{ test_equality_complexity

@pytest.mark.xfail
def test_equality_complexity():
    # NOTE: https://github.com/inducer/pymbolic/issues/73
    from numpy.random import default_rng

    def construct_intestine_graph(depth=64, seed=0):
        rng = default_rng(seed)
        x = prim.Variable("x")

        for _ in range(depth):
            coeff1, coeff2 = rng.integers(1, 10, 2)
            x = coeff1 * x + coeff2 * x

        return x

    def check_equality():
        graph1 = construct_intestine_graph()
        graph2 = construct_intestine_graph()
        graph3 = construct_intestine_graph(seed=3)

        assert graph1 == graph2
        assert graph2 == graph1
        assert graph1 != graph3
        assert graph2 != graph3

    # NOTE: this should finish in a second!
    import multiprocessing
    p = multiprocessing.Process(target=check_equality)
    p.start()
    p.join(timeout=1)

    is_alive = p.is_alive()
    if p.is_alive():
        p.terminate()

    assert not is_alive

# }}}


# {{{ test_cached_mapper_memoizes

class InCacheVerifier(WalkMapper):
    def __init__(self, cached_mapper, walk_call_functions=True):
        super().__init__()
        self.cached_mapper = cached_mapper
        self.walk_call_functions = walk_call_functions

    def post_visit(self, expr):
        if isinstance(expr, prim.ExpressionNode):
            assert (self.cached_mapper.get_cache_key(expr)
                    in self.cached_mapper._cache)

    def map_call(self, expr):
        if not self.visit(expr):
            return

        if self.walk_call_functions:
            self.rec(expr.function)

        for child in expr.parameters:
            self.rec(child)

        self.post_visit(expr)


def test_cached_mapper_memoizes():
    from testlib import (
        AlwaysFlatteningCachedIdentityMapper,
        AlwaysFlatteningIdentityMapper,
    )
    ntests = 40
    for i in range(ntests):
        expr = generate_random_expression(seed=(5+i))

        # {{{ always flattening identity mapper

        # Note: Prefer AlwaysFlatteningIdentityMapper over IdentityMapper as
        # the flattening logic in IdentityMapper checks for identity across
        # traversal results => leading to discrepancy b/w
        # 'CachedIdentityMapper' and 'IdentityMapper'

        cached_mapper = AlwaysFlatteningCachedIdentityMapper()
        uncached_mapper = AlwaysFlatteningIdentityMapper()
        assert uncached_mapper(expr) == cached_mapper(expr)
        verifier = InCacheVerifier(cached_mapper)
        verifier(expr)

        # }}}

        # {{{ dependency mapper

        mapper = DependencyMapper(include_calls="descend_args")
        cached_mapper = CachedDependencyMapper(include_calls="descend_args")
        assert cached_mapper(expr) == mapper(expr)
        verifier = InCacheVerifier(cached_mapper,
                                   # dep. mapper does not go over functions
                                   walk_call_functions=False
                                   )
        verifier(expr)

        # }}}


def test_cached_mapper_differentiates_float_int():
    # pymbolic.git<=d343cf14 failed this regression.
    from pymbolic.mapper import CachedIdentityMapper
    expr = prim.Sum((4, 4.0))
    cached_mapper = CachedIdentityMapper()
    new_expr = cached_mapper(expr)
    assert isinstance(new_expr.children[0], int)
    assert isinstance(new_expr.children[1], float)

# }}}


# {{{ test_mapper_optimizer

def test_mapper_optimizer():
    from testlib import BIG_EXPR_STR, OptimizedRenamer, Renamer

    from pymbolic.mapper import CachedIdentityMapper

    expr = parse(BIG_EXPR_STR)
    expr = CachedIdentityMapper()(expr)  # remove duplicate nodes

    result_ref = Renamer()(expr)
    result_opt = OptimizedRenamer()(expr)

    assert result_ref == result_opt

# }}}


def test_nodecount():
    from pymbolic.mapper.analysis import get_num_nodes
    expr = prim.Sum((4, 4.0))

    assert get_num_nodes(expr) == 3

    x = prim.Variable("x")
    y = prim.Variable("y")
    z = prim.Variable("z")

    subexpr = prim.make_common_subexpression(4 * (x**2 + y + z))
    expr = 3*subexpr + subexpr + subexpr + subexpr
    expr = expr + expr + expr

    assert get_num_nodes(expr) == 12


def test_python_ast_interop_roundtrip():
    from pymbolic.interop.ast import ASTToPymbolic, PymbolicToASTMapper

    ast2p = ASTToPymbolic()
    p2ast = PymbolicToASTMapper()
    ntests = 40
    for i in range(ntests):
        expr = generate_random_expression(seed=(5+i))
        assert ast2p(p2ast(expr)) == expr


# {{{ test derived stringifiers

@prim.expr_dataclass()
class CustomOperator:
    child: Expression

    def make_stringifier(self, originating_stringifier=None):
        return OperatorStringifier()


class OperatorStringifier(StringifyMapper[[]]):
    def map_custom_operator(self, expr: CustomOperator):
        return f"Op({self.rec(expr.child)})"


def test_derived_stringifier() -> None:
    str(CustomOperator(5))

# }}}


# {{{ test_flatten

class IntegerFlattenMapper(FlattenMapper):
    def is_expr_integer_valued(self, expr: Expression) -> bool:
        return True


def test_flatten():
    expr = parse("(3 + x) % 1")

    assert IntegerFlattenMapper()(expr) != expr
    assert FlattenMapper()(expr) == expr

    assert evaluate_kw(IntegerFlattenMapper()(expr), x=1) == 0
    assert abs(evaluate_kw(FlattenMapper()(expr), x=1.1) - 0.1) < 1e-12

    expr = parse("(3 + x) // 1")

    assert IntegerFlattenMapper()(expr) != expr
    assert FlattenMapper()(expr) == expr

    assert evaluate_kw(IntegerFlattenMapper()(expr), x=1) == 4
    assert abs(evaluate_kw(FlattenMapper()(expr), x=1.1) - 4) < 1e-12

# }}}


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])

# vim: fdm=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from collections.abc import Sequence
from typing import TYPE_CHECKING, ClassVar, Concatenate
from warnings import warn

from typing_extensions import deprecated

import pymbolic.primitives as p
from pymbolic.mapper import CachedMapper, Mapper, P
from pymbolic.typing import Expression


if TYPE_CHECKING:
    import numpy as np

    from pymbolic.geometric_algebra import MultiVector


__doc__ = """
.. _prec-constants:

Precedence constants
********************

.. data:: PREC_CALL
.. data:: PREC_POWER
.. data:: PREC_UNARY
.. data:: PREC_PRODUCT
.. data:: PREC_SUM
.. data:: PREC_SHIFT
.. data:: PREC_BITWISE_AND
.. data:: PREC_BITWISE_XOR
.. data:: PREC_BITWISE_OR
.. data:: PREC_COMPARISON
.. data:: PREC_LOGICAL_AND
.. data:: PREC_LOGICAL_OR
.. data:: PREC_IF
.. data:: PREC_NONE

Mappers
*******

.. autoclass:: StringifyMapper

    .. automethod:: __call__

.. autoclass:: CSESplittingStringifyMapperMixin

.. autoclass:: LaTeXMapper
"""


PREC_CALL = 15
PREC_POWER = 14
PREC_UNARY = 13
PREC_PRODUCT = 12
PREC_SUM = 11
PREC_SHIFT = 10
PREC_BITWISE_AND = 9
PREC_BITWISE_XOR = 8
PREC_BITWISE_OR = 7
PREC_COMPARISON = 6
PREC_LOGICAL_AND = 5
PREC_LOGICAL_OR = 4
PREC_IF = 3
PREC_NONE = 0


# {{{ stringifier


class StringifyMapper(Mapper[str, Concatenate[int, P]]):
    """A mapper to turn an expression tree into a string.

    :class:`pymbolic.ExpressionNode.__str__` is often implemented using
    this mapper.

    When it encounters an unsupported :class:`pymbolic.ExpressionNode`
    subclass, it calls its :meth:`pymbolic.ExpressionNode.make_stringifier`
    method to get a :class:`StringifyMapper` that potentially does.
    """

    # {{{ replaceable string composition interface

    def format(self, s: str, *args: object) -> str:
        return s % args

    def join(self, joiner: str, seq: Sequence[Expression]) -> str:
        return self.format(joiner.join("%s" for _ in seq), *seq)

    # {{{ deprecated junk

    @deprecated("interface not type-safe, use rec_with_parens_around_types")
    def rec_with_force_parens_around(self, expr, *args, **kwargs):
        warn(
            "rec_with_force_parens_around is deprecated and will be removed in 2025. "
            "Use rec_with_parens_around_types instead. ",
            DeprecationWarning,
            stacklevel=2,
        )
        # Not currently possible to make this type-safe:
        # https://peps.python.org/pep-0612/#concatenating-keyword-parameters

        force_parens_around = kwargs.pop("force_parens_around", ())

        result = self.rec(expr, *args, **kwargs)

        if isinstance(expr, force_parens_around):
            result = f"({result})"

        return result

    def join_rec(
        self,
        joiner: str,
        seq: Sequence[Expression],
        prec: int,
        *args,
        **kwargs,  # force_with_parens_around may hide in here
    ) -> str:
        f = joiner.join("%s" for _ in seq)

        if "force_parens_around" in kwargs:
            warn(
                "Passing force_parens_around join_rec is deprecated and will be "
                "removed in 2025. "
                "Use join_rec_with_parens_around_types instead. ",
                DeprecationWarning,
                stacklevel=2,
            )
            # Not currently possible to make this type-safe:
            # https://peps.python.org/pep-0612/#concatenating-keyword-parameters
            parens_around_types: tuple[type, ...] = kwargs.pop("force_parens_around")
            return self.join_rec_with_parens_around_types(
                joiner, seq, prec, parens_around_types, *args, **kwargs
            )

        return self.format(
            f,
            *[self.rec(i, prec, *args, **kwargs) for i in seq],
        )

    # }}}

    def rec_with_parens_around_types(
        self,
        expr: Expression,
        enclosing_prec: int,
        parens_around: tuple[type, ...],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        result = self.rec(expr, enclosing_prec, *args, **kwargs)

        if isinstance(expr, parens_around):
            result = f"({result})"

        return result

    def join_rec_with_parens_around_types(
        self,
        joiner: str,
        seq: Sequence[Expression],
        prec: int,
        parens_around_types: tuple[type, ...],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        f = joiner.join("%s" for _ in seq)
        return self.format(
            f,
            *[
                self.rec_with_parens_around_types(
                    i, prec, parens_around_types, *args, **kwargs
                )
                for i in seq
            ],
        )

    def parenthesize(self, s: str) -> str:
        return f"({s})"

    def parenthesize_if_needed(self, s: str, enclosing_prec: int, my_prec: int) -> str:
        if enclosing_prec > my_prec:
            return f"({s})"
        else:
            return s

    # }}}

    # {{{ mappings

    def handle_unsupported_expression(
        self, expr, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        strifier = expr.make_stringifier(self)
        if isinstance(self, type(strifier)):
            raise ValueError(f"stringifier '{self}' can't handle '{expr.__class__}'")
        return strifier(expr, enclosing_prec, *args, **kwargs)

    def map_constant(
        self, expr: object, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        result = str(expr)

        if (
            not (result.startswith("(") and result.endswith(")"))
            and ("-" in result or "+" in result)
            and (enclosing_prec > PREC_SUM)
        ):
            return self.parenthesize(result)
        else:
            return result

    def map_variable(
        self, expr: p.Variable, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return expr.name

    def map_wildcard(
        self, expr: p.Wildcard, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return "*"

    def map_function_symbol(
        self,
        expr: p.FunctionSymbol,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return expr.__class__.__name__

    def map_call(
        self, expr: p.Call, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.format(
            "%s(%s)",
            self.rec(expr.function, PREC_CALL, *args, **kwargs),
            self.join_rec(", ", expr.parameters, PREC_NONE, *args, **kwargs),
        )

    def map_call_with_kwargs(
        self,
        expr: p.CallWithKwargs,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        args_strings = tuple([
            self.rec(ch, PREC_NONE, *args, **kwargs) for ch in expr.parameters
        ]) + tuple([
            "{}={}".format(name, self.rec(ch, PREC_NONE, *args, **kwargs))
            for name, ch in expr.kw_parameters.items()
        ])
        return self.format(
            "%s(%s)",
            self.rec(expr.function, PREC_CALL, *args, **kwargs),
            ", ".join(args_strings),
        )

    def map_subscript(
        self, expr: p.Subscript, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        if isinstance(expr.index, tuple):
            index_str = self.join_rec(", ", expr.index, PREC_NONE, *args, **kwargs)
        else:
            index_str = self.rec(expr.index, PREC_NONE, *args, **kwargs)

        return self.parenthesize_if_needed(
            self.format(
                "%s[%s]",
                self.rec(expr.aggregate, PREC_CALL, *args, **kwargs),
                index_str,
            ),
            enclosing_prec,
            PREC_CALL,
        )

    def map_lookup(
        self, expr: p.Lookup, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.format(
                "%s.%s", self.rec(expr.aggregate, PREC_CALL, *args, **kwargs), expr.name
            ),
            enclosing_prec,
            PREC_CALL,
        )

    def map_sum(
        self, expr: p.Sum, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.join_rec(" + ", expr.children, PREC_SUM, *args, **kwargs),
            enclosing_prec,
            PREC_SUM,
        )

    # {{{ multiplicative operators

    multiplicative_primitives = (p.Product, p.Quotient, p.FloorDiv, p.Remainder)

    def map_product(
        self, expr: p.Product, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.join_rec_with_parens_around_types(
                "*",
                expr.children,
                PREC_PRODUCT,
                (p.Quotient, p.FloorDiv, p.Remainder),
                *args,
                **kwargs,
            ),
            enclosing_prec,
            PREC_PRODUCT,
        )

    def map_quotient(
        self, expr: p.Quotient, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.format(
                "%s / %s",
                # space is necessary--otherwise '/*' becomes
                # start-of-comment in C. ('*' from dereference)
                self.rec_with_parens_around_types(
                    expr.numerator,
                    PREC_PRODUCT,
                    self.multiplicative_primitives,
                    *args,
                    **kwargs,
                ),
                self.rec_with_parens_around_types(
                    expr.denominator,
                    PREC_PRODUCT,
                    self.multiplicative_primitives,
                    *args,
                    **kwargs,
                ),
            ),
            enclosing_prec,
            PREC_PRODUCT,
        )

    def map_floor_div(
        self, expr: p.FloorDiv, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.format(
                "%s // %s",
                self.rec_with_parens_around_types(
                    expr.numerator,
                    PREC_PRODUCT,
                    self.multiplicative_primitives,
                    *args,
                    **kwargs,
                ),
                self.rec_with_parens_around_types(
                    expr.denominator,
                    PREC_PRODUCT,
                    self.multiplicative_primitives,
                    *args,
                    **kwargs,
                ),
            ),
            enclosing_prec,
            PREC_PRODUCT,
        )

    def map_remainder(
        self, expr: p.Remainder, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.format(
                "%s %% %s",
                self.rec_with_parens_around_types(
                    expr.numerator,
                    PREC_PRODUCT,
                    self.multiplicative_primitives,
                    *args,
                    **kwargs,
                ),
                self.rec_with_parens_around_types(
                    expr.denominator,
                    PREC_PRODUCT,
                    self.multiplicative_primitives,
                    *args,
                    **kwargs,
                ),
            ),
            enclosing_prec,
            PREC_PRODUCT,
        )

    # }}}

    def map_power(
        self, expr: p.Power, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.format(
                "%s**%s",
                self.rec(expr.base, PREC_POWER, *args, **kwargs),
                self.rec(expr.exponent, PREC_POWER, *args, **kwargs),
            ),
            enclosing_prec,
            PREC_POWER,
        )

    def map_left_shift(
        self, expr: p.LeftShift, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            # +1 to address
            # https://gitlab.tiker.net/inducer/pymbolic/issues/6
            self.format(
                "%s << %s",
                self.rec(expr.shiftee, PREC_SHIFT + 1, *args, **kwargs),
                self.rec(expr.shift, PREC_SHIFT + 1, *args, **kwargs),
            ),
            enclosing_prec,
            PREC_SHIFT,
        )

    def map_right_shift(
        self,
        expr: p.RightShift,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return self.parenthesize_if_needed(
            # +1 to address
            # https://gitlab.tiker.net/inducer/pymbolic/issues/6
            self.format(
                "%s >> %s",
                self.rec(expr.shiftee, PREC_SHIFT + 1, *args, **kwargs),
                self.rec(expr.shift, PREC_SHIFT + 1, *args, **kwargs),
            ),
            enclosing_prec,
            PREC_SHIFT,
        )

    def map_bitwise_not(
        self,
        expr: p.BitwiseNot,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return self.parenthesize_if_needed(
            "~" + self.rec(expr.child, PREC_UNARY, *args, **kwargs),
            enclosing_prec,
            PREC_UNARY,
        )

    def map_bitwise_or(
        self, expr: p.BitwiseOr, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.join_rec(" | ", expr.children, PREC_BITWISE_OR, *args, **kwargs),
            enclosing_prec,
            PREC_BITWISE_OR,
        )

    def map_bitwise_xor(
        self,
        expr: p.BitwiseXor,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return self.parenthesize_if_needed(
            self.join_rec(" ^ ", expr.children, PREC_BITWISE_XOR, *args, **kwargs),
            enclosing_prec,
            PREC_BITWISE_XOR,
        )

    def map_bitwise_and(
        self,
        expr: p.BitwiseAnd,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return self.parenthesize_if_needed(
            self.join_rec(" & ", expr.children, PREC_BITWISE_AND, *args, **kwargs),
            enclosing_prec,
            PREC_BITWISE_AND,
        )

    def map_comparison(
        self, expr: p.Comparison, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.format(
                "%s %s %s",
                self.rec(expr.left, PREC_COMPARISON, *args, **kwargs),
                expr.operator,
                self.rec(expr.right, PREC_COMPARISON, *args, **kwargs),
            ),
            enclosing_prec,
            PREC_COMPARISON,
        )

    def map_logical_not(
        self,
        expr: p.LogicalNot,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return self.parenthesize_if_needed(
            "not " + self.rec(expr.child, PREC_UNARY, *args, **kwargs),
            enclosing_prec,
            PREC_UNARY,
        )

    def map_logical_or(
        self, expr: p.LogicalOr, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.join_rec(" or ", expr.children, PREC_LOGICAL_OR, *args, **kwargs),
            enclosing_prec,
            PREC_LOGICAL_OR,
        )

    def map_logical_and(
        self,
        expr: p.LogicalAnd,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return self.parenthesize_if_needed(
            self.join_rec(" and ", expr.children, PREC_LOGICAL_AND, *args, **kwargs),
            enclosing_prec,
            PREC_LOGICAL_AND,
        )

    def map_list(
        self,
        expr: list[Expression],
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return self.format(
            "[%s]", self.join_rec(", ", expr, PREC_NONE, *args, **kwargs)
        )

    map_vector = map_list

    def map_tuple(
        self,
        expr: tuple[Expression, ...],
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        el_str = ", ".join(
            self.rec(child, PREC_NONE, *args, **kwargs) for child in expr
        )
        if len(expr) == 1:
            el_str += ","

        return f"({el_str})"

    def map_numpy_array(
        self,
        expr: np.ndarray,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        import numpy

        str_array = numpy.zeros(expr.shape, dtype="object")
        max_length = 0
        for i in numpy.ndindex(expr.shape):
            s = self.rec(expr[i], PREC_NONE, *args, **kwargs)
            max_length = max(len(s), max_length)
            str_array[i] = s.replace("\n", "\n  ")

        if len(expr.shape) == 1 and max_length < 15:
            return "array({})".format(", ".join(str_array))
        else:
            lines = [
                "  {}: {}\n".format(",".join(str(i_i) for i_i in i), val)
                for i, val in numpy.ndenumerate(str_array)
            ]
            if max_length > 70:
                splitter = "  " + "-" * 75 + "\n"
                return "array(\n{})".format(splitter.join(lines))
            else:
                return "array(\n{})".format("".join(lines))

    def map_multivector(
        self,
        expr: MultiVector,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return expr.stringify(self.rec, enclosing_prec, *args, **kwargs)

    def map_common_subexpression(
        self,
        expr: p.CommonSubexpression,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        from pymbolic.primitives import CommonSubexpression

        if type(expr) is CommonSubexpression:
            type_name = "CSE"
        else:
            type_name = type(expr).__name__

        return self.format(
            "%s(%s)", type_name, self.rec(expr.child, PREC_NONE, *args, **kwargs)
        )

    def map_if(
        self, expr: p.If, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            "{} if {} else {}".format(
                self.rec(expr.then, PREC_LOGICAL_OR, *args, **kwargs),
                self.rec(expr.condition, PREC_LOGICAL_OR, *args, **kwargs),
                self.rec(expr.else_, PREC_LOGICAL_OR, *args, **kwargs),
            ),
            enclosing_prec,
            PREC_IF,
        )

    def map_min(
        self, expr: p.Min, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        what = type(expr).__name__.lower()
        return self.format(
            "%s(%s)",
            what,
            self.join_rec(", ", expr.children, PREC_NONE, *args, **kwargs),
        )

    def map_max(
        self, expr: p.Max, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        what = type(expr).__name__.lower()
        return self.format(
            "%s(%s)",
            what,
            self.join_rec(", ", expr.children, PREC_NONE, *args, **kwargs),
        )

    def map_derivative(
        self, expr: p.Derivative, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        derivs = " ".join(f"d/d{v}" for v in expr.variables)

        return "{} {}".format(
            derivs, self.rec(expr.child, PREC_PRODUCT, *args, **kwargs)
        )

    def map_substitution(
        self,
        expr: p.Substitution,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        substs = ", ".join(
            "{}={}".format(name, self.rec(val, PREC_NONE, *args, **kwargs))
            for name, val in zip(expr.variables, expr.values, strict=True)
        )

        return "[%s]{%s}" % (self.rec(expr.child, PREC_NONE, *args, **kwargs), substs)

    def map_slice(
        self, expr: p.Slice, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        children = []
        for child in expr.children:
            if child is None:
                children.append("")
            else:
                children.append(self.rec(child, PREC_NONE, *args, **kwargs))

        return self.parenthesize_if_needed(
            ":".join(children), enclosing_prec, PREC_NONE
        )

    def map_nan(
        self, expr: p.NaN, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return "NaN"

    # }}}

    def __call__(self, expr, prec=PREC_NONE, *args: P.args, **kwargs: P.kwargs) -> str:
        """Return a string corresponding to *expr*. If the enclosing
        precedence level *prec* is higher than *prec* (see :ref:`prec-constants`),
        parenthesize the result.
        """

        return Mapper.__call__(self, expr, prec, *args, **kwargs)


class CachedStringifyMapper(StringifyMapper, CachedMapper):
    def __init__(self) -> None:
        StringifyMapper.__init__(self)
        CachedMapper.__init__(self)

    def __call__(self, expr, prec=PREC_NONE, *args: P.args, **kwargs: P.kwargs) -> str:
        return CachedMapper.__call__(expr, prec, *args, **kwargs)


# }}}


# {{{ cse-splitting stringifier


class CSESplittingStringifyMapperMixin(Mapper[str, Concatenate[int, P]]):
    """A :term:`mix-in` for subclasses of
    :class:`StringifyMapper` that collects
    "variable assignments" for
    :class:`pymbolic.primitives.CommonSubexpression` objects.

    .. attribute:: cse_to_name

        A :class:`dict` mapping expressions to CSE variable names.

    .. attribute:: cse_names

        A :class:`set` of names already assigned.

    .. attribute:: cse_name_list

        A :class:`list` of tuples of names and their string representations,
        in order of their dependencies. When generating code, walk down these names
        in order, and the generated code will never reference
        an undefined variable.

    See :class:`pymbolic.mapper.c_code.CCodeMapper` for an example
    of the use of this mix-in.
    """

    cse_to_name: dict[Expression, str]
    cse_names: set[str]
    cse_name_list: list[tuple[str, str]]

    def __init__(self) -> None:
        self.cse_to_name = {}
        self.cse_names = set()
        self.cse_name_list = []

        super().__init__()

    def map_common_subexpression(
        self,
        expr: p.CommonSubexpression,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        try:
            cse_name = self.cse_to_name[expr.child]
        except KeyError:
            str_child = self.rec(expr.child, PREC_NONE, *args, **kwargs)

            if expr.prefix is not None:

                def generate_cse_names():
                    yield expr.prefix
                    i = 2
                    while True:
                        yield expr.prefix + f"_{i}"
                        i += 1

            else:

                def generate_cse_names():
                    i = 0
                    while True:
                        yield "CSE" + str(i)
                        i += 1

            for cse_name in generate_cse_names():
                if cse_name not in self.cse_names:
                    break

            self.cse_name_list.append((cse_name, str_child))
            self.cse_to_name[expr.child] = cse_name
            self.cse_names.add(cse_name)

        return cse_name

    def get_cse_strings(self) -> list[str]:
        return [
            f"{cse_name} : {cse_str}"
            for cse_name, cse_str in sorted(getattr(self, "cse_name_list", []))
        ]


# }}}


# {{{ sorting stringifier


class SortingStringifyMapper(StringifyMapper[P]):
    def __init__(self, reverse=True):
        super().__init__()
        self.reverse = reverse

    def map_sum(
        self, expr: p.Sum, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        entries = [self.rec(i, PREC_SUM, *args, **kwargs) for i in expr.children]
        entries.sort(reverse=self.reverse)
        return self.parenthesize_if_needed("+".join(entries), enclosing_prec, PREC_SUM)

    def map_product(
        self, expr: p.Product, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        entries = [self.rec(i, PREC_PRODUCT, *args, **kwargs) for i in expr.children]
        entries.sort(reverse=self.reverse)
        return self.parenthesize_if_needed(
            "*".join(entries), enclosing_prec, PREC_PRODUCT
        )


# }}}


# {{{ simplifying, sorting stringifier


class SimplifyingSortingStringifyMapper(StringifyMapper):
    def __init__(self, reverse=True):
        super().__init__()
        self.reverse = reverse

    def map_sum(
        self, expr: p.Sum, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        def get_neg_product(expr):
            from pymbolic.primitives import Product, is_zero

            if (
                isinstance(expr, Product)
                and len(expr.children)
                and is_zero(expr.children[0] + 1)
            ):
                if len(expr.children) == 2:
                    # only the minus sign and the other child
                    return expr.children[1]
                else:
                    return Product(expr.children[1:])
            else:
                return None

        positives = []
        negatives = []

        for ch in expr.children:
            neg_prod = get_neg_product(ch)
            if neg_prod is not None:
                negatives.append(self.rec(neg_prod, PREC_PRODUCT, *args, **kwargs))
            else:
                positives.append(self.rec(ch, PREC_SUM, *args, **kwargs))

        positives.sort(reverse=self.reverse)
        positives_str = " + ".join(positives)
        negatives.sort(reverse=self.reverse)
        negatives_str = "".join(self.format(" - %s", entry) for entry in negatives)

        result = positives_str + negatives_str

        return self.parenthesize_if_needed(result, enclosing_prec, PREC_SUM)

    def map_product(
        self, expr: p.Product, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        entries = []
        i = 0
        from pymbolic.primitives import is_zero

        while i < len(expr.children):
            child = expr.children[i]
            if False and is_zero(child + 1) and i + 1 < len(expr.children):
                # NOTE: That space needs to be there.
                # Otherwise two unary minus signs merge into a pre-decrement.
                entries.append(
                    self.format(
                        "- %s",
                        self.rec(expr.children[i + 1], PREC_UNARY, *args, **kwargs),
                    )
                )
                i += 2
            else:
                entries.append(self.rec(child, PREC_PRODUCT, *args, **kwargs))
                i += 1

        entries.sort(reverse=self.reverse)
        result = "*".join(entries)

        return self.parenthesize_if_needed(result, enclosing_prec, PREC_PRODUCT)


# }}}


# {{{ latex stringifier


class LaTeXMapper(StringifyMapper):
    COMPARISON_OP_TO_LATEX: ClassVar[dict[str, str]] = {
        "==": r"=",
        "!=": r"\ne",
        "<=": r"\le",
        ">=": r"\ge",
        "<": r"<",
        ">": r">",
    }

    def map_remainder(
        self, expr: p.Remainder, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.format(
            r"(%s \bmod %s)",
            self.rec(expr.numerator, PREC_PRODUCT, *args, **kwargs),
            self.rec(expr.denominator, PREC_POWER, *args, **kwargs),
        )

    def map_left_shift(
        self, expr: p.LeftShift, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.format(
                r"%s \ll %s",
                self.rec(expr.shiftee, PREC_SHIFT + 1, *args, **kwargs),
                self.rec(expr.shift, PREC_SHIFT + 1, *args, **kwargs),
            ),
            enclosing_prec,
            PREC_SHIFT,
        )

    def map_right_shift(
        self,
        expr: p.RightShift,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return self.parenthesize_if_needed(
            self.format(
                r"%s \gg %s",
                self.rec(expr.shiftee, PREC_SHIFT + 1, *args, **kwargs),
                self.rec(expr.shift, PREC_SHIFT + 1, *args, **kwargs),
            ),
            enclosing_prec,
            PREC_SHIFT,
        )

    def map_bitwise_xor(
        self,
        expr: p.BitwiseXor,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return self.parenthesize_if_needed(
            self.join_rec(
                r" \wedge ", expr.children, PREC_BITWISE_XOR, *args, **kwargs
            ),
            enclosing_prec,
            PREC_BITWISE_XOR,
        )

    def map_product(
        self, expr: p.Product, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.join_rec(" ", expr.children, PREC_PRODUCT, *args, **kwargs),
            enclosing_prec,
            PREC_PRODUCT,
        )

    def map_power(
        self, expr: p.Power, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.format(
                "{%s}^{%s}",
                self.rec(expr.base, PREC_NONE, *args, **kwargs),
                self.rec(expr.exponent, PREC_NONE, *args, **kwargs),
            ),
            enclosing_prec,
            PREC_NONE,
        )

    def map_min(
        self, expr: p.Min, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        from pytools import is_single_valued

        if is_single_valued(expr.children):
            return self.rec(expr.children[0], enclosing_prec)

        what = type(expr).__name__.lower()
        return self.format(
            r"\%s(%s)",
            what,
            self.join_rec(", ", expr.children, PREC_NONE, *args, **kwargs),
        )

    def map_max(
        self, expr: p.Max, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        from pytools import is_single_valued

        if is_single_valued(expr.children):
            return self.rec(expr.children[0], enclosing_prec)

        what = type(expr).__name__.lower()
        return self.format(
            r"\%s(%s)",
            what,
            self.join_rec(", ", expr.children, PREC_NONE, *args, **kwargs),
        )

    def map_floor_div(
        self, expr: p.FloorDiv, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.format(
            r"\lfloor {%s} / {%s} \rfloor",
            self.rec(expr.numerator, PREC_NONE, *args, **kwargs),
            self.rec(expr.denominator, PREC_NONE, *args, **kwargs),
        )

    def map_subscript(
        self, expr: p.Subscript, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        if isinstance(expr.index, tuple):
            index_str = self.join_rec(", ", expr.index, PREC_NONE, *args, **kwargs)
        else:
            index_str = self.rec(expr.index, PREC_NONE, *args, **kwargs)

        return self.format(
            "{%s}_{%s}", self.rec(expr.aggregate, PREC_CALL, *args, **kwargs), index_str
        )

    def map_logical_not(
        self,
        expr: p.LogicalNot,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return self.parenthesize_if_needed(
            r"\neg " + self.rec(expr.child, PREC_UNARY, *args, **kwargs),
            enclosing_prec,
            PREC_UNARY,
        )

    def map_logical_or(
        self, expr: p.LogicalOr, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.join_rec(r" \vee ", expr.children, PREC_LOGICAL_OR, *args, **kwargs),
            enclosing_prec,
            PREC_LOGICAL_OR,
        )

    def map_logical_and(
        self,
        expr: p.LogicalAnd,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        return self.parenthesize_if_needed(
            self.join_rec(
                r" \wedge ", expr.children, PREC_LOGICAL_AND, *args, **kwargs
            ),
            enclosing_prec,
            PREC_LOGICAL_AND,
        )

    def map_comparison(
        self, expr: p.Comparison, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        return self.parenthesize_if_needed(
            self.format(
                "%s %s %s",
                self.rec(expr.left, PREC_COMPARISON, *args, **kwargs),
                self.COMPARISON_OP_TO_LATEX[expr.operator],
                self.rec(expr.right, PREC_COMPARISON, *args, **kwargs),
            ),
            enclosing_prec,
            PREC_COMPARISON,
        )

    def map_substitution(
        self,
        expr: p.Substitution,
        enclosing_prec: int,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        substs = ", ".join(
            "{}={}".format(name, self.rec(val, PREC_NONE, *args, **kwargs))
            for name, val in zip(expr.variables, expr.values, strict=True)
        )

        return self.format(
            r"[%s]\{%s\}", self.rec(expr.child, PREC_NONE, *args, **kwargs), substs
        )

    def map_derivative(
        self, expr: p.Derivative, enclosing_prec: int, *args: P.args, **kwargs: P.kwargs
    ) -> str:
        derivs = " ".join(r"\frac{\partial}{\partial %s}" % v for v in expr.variables)

        return self.format(
            "%s %s", derivs, self.rec(expr.child, PREC_PRODUCT, *args, **kwargs)
        )


# }}}

# vim: fdm=marker



from __future__ import annotations


__copyright__ = """
Copyright (C) 2017 Matt Wala
Copyright (C) 2009-2013 Andreas Kloeckner
"""

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import sympy

import pymbolic.primitives as prim
from pymbolic.interop.common import PymbolicToSympyLikeMapper, SympyLikeToPymbolicMapper


__doc__ = """
.. class:: SympyToPymbolicMapper

    .. method:: __call__(expr)

.. class:: PymbolicToSympyMapper

    .. method:: __call__(expr)
"""


# {{{ sympy -> pymbolic

class SympyToPymbolicMapper(SympyLikeToPymbolicMapper):

    def map_ImaginaryUnit(self, expr):  # noqa
        return 1j

    map_Float = SympyLikeToPymbolicMapper.to_float  # noqa: N815

    map_NumberSymbol = SympyLikeToPymbolicMapper.to_float  # noqa: N815

    def function_name(self, expr):
        return type(expr).__name__

    # only called for Py2
    def map_long(self, expr):
        return long(expr)  # noqa  pylint:disable=undefined-variable

    def map_Indexed(self, expr):  # noqa
        if len(expr.args) == 2:
            return prim.Subscript(
                self.rec(expr.args[0].args[0]),
                self.rec(expr.args[1]),
                )

        return prim.Subscript(
            self.rec(expr.args[0].args[0]),
            tuple([self.rec(i) for i in expr.args[1:]])
            )

    def map_CSE(self, expr):  # noqa
        return prim.CommonSubexpression(
                self.rec(expr.args[0]), expr.prefix, expr.scope)

    def map_Piecewise(self, expr):  # noqa
        # We only handle piecewises with 2 arguments!
        if not len(expr.args) == 2:
            raise NotImplementedError
        # We only handle if/else cases
        if not (expr.args[1][1].is_Boolean and bool(expr.args[1][1]) is True):
            raise NotImplementedError
        then = self.rec(expr.args[0][0])
        else_ = self.rec(expr.args[1][0])
        cond = self.rec(expr.args[0][1])
        return prim.If(cond, then, else_)

# }}}


# {{{ pymbolic -> sympy

class PymbolicToSympyMapper(PymbolicToSympyLikeMapper):

    sym = sympy

    def raise_conversion_error(self, expr):
        raise RuntimeError(f"do not know how to translate '{expr}' to sympy")

    def map_subscript(self, expr):
        return self.sym.Indexed(
            self.sym.IndexedBase(self.rec(expr.aggregate)),
            *[self.rec(i) for i in expr.index_tuple]
            )
# }}}


class CSE(sympy.Function):
    """A function to translate to a Pymbolic CSE."""
    nargs = 1


def make_cse(arg, prefix=None, scope=None):
    result = CSE(arg)
    result.prefix = prefix
    result.scope = scope
    return result


# vim: fdm=marker



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

import multiset
import numpy as np
from matchpy import Expression as MatchpyExpression

import pymbolic.interop.matchpy as m
import pymbolic.primitives as p
from pymbolic.interop.matchpy.mapper import Mapper as BaseMatchPyMapper
from pymbolic.mapper import Mapper as BasePymMapper
from pymbolic.typing import Scalar as PbScalar


# {{{ to matchpy

class ToMatchpyExpressionMapper(BasePymMapper):
    """
    Mapper to convert instances of :class:`pymbolic.primitives.Expression` to
    :class:`pymbolic.interop.matchpy.PymbolicOperation`.
    """
    def map_constant(self, expr: Any) -> m.Scalar:
        if np.isscalar(expr):
            return m.Scalar(expr)

        raise NotImplementedError(expr)

    def map_variable(self, expr: p.Variable) -> m.Variable:
        return m.Variable(m.Id(expr.name))

    def map_call(self, expr: p.Call) -> m.Call:
        return m.Call(self.rec(expr.function),
                      m.TupleOp(tuple(self.rec(p)
                                      for p in expr.parameters)))

    def map_subscript(self, expr: p.Subscript) -> m.Subscript:
        return m.Subscript(self.rec(expr.aggregate),
                           m.TupleOp(tuple(self.rec(idx)
                                           for idx in expr.index_tuple)))

    def map_sum(self, expr: p.Sum) -> m.Sum:
        return m.Sum(*[self.rec(child)
                       for child in expr.children])

    def map_product(self, expr: p.Product) -> m.Product:
        return m.Product(*[self.rec(child)
                           for child in expr.children])

    def map_quotient(self, expr: p.Quotient) -> m.TrueDiv:
        return m.TrueDiv(self.rec(expr.numerator), self.rec(expr.denominator))

    def map_floor_div(self, expr: p.FloorDiv) -> m.FloorDiv:
        return m.FloorDiv(self.rec(expr.numerator), self.rec(expr.denominator))

    def map_remainder(self, expr: p.Remainder) -> m.Modulo:
        return m.Modulo(self.rec(expr.numerator), self.rec(expr.denominator))

    def map_power(self, expr: p.Power) -> m.Power:
        return m.Power(self.rec(expr.base), self.rec(expr.exponent))

    def map_left_shift(self, expr: p.LeftShift) -> m.LeftShift:
        return m.LeftShift(self.rec(expr.shiftee), self.rec(expr.shift))

    def map_right_shift(self, expr: p.RightShift) -> m.RightShift:
        return m.RightShift(self.rec(expr.shiftee), self.rec(expr.shift))

    def map_bitwise_not(self, expr: p.BitwiseNot) -> m.BitwiseNot:
        return m.BitwiseNot(self.rec(expr.child))

    def map_bitwise_or(self, expr: p.BitwiseOr) -> m.BitwiseOr:
        return m.BitwiseOr(*[self.rec(child)
                             for child in expr.children])

    def map_bitwise_and(self, expr: p.BitwiseAnd) -> m.BitwiseAnd:
        return m.BitwiseAnd(*[self.rec(child)
                              for child in expr.children])

    def map_bitwise_xor(self, expr: p.BitwiseXor) -> m.BitwiseXor:
        return m.BitwiseXor(*[self.rec(child)
                              for child in expr.children])

    def map_logical_not(self, expr: p.LogicalNot) -> m.LogicalNot:
        return m.LogicalNot(self.rec(expr.child))

    def map_logical_or(self, expr: p.LogicalOr) -> m.LogicalOr:
        return m.LogicalOr(*[self.rec(child)
                             for child in expr.children])

    def map_logical_and(self, expr: p.LogicalAnd) -> m.LogicalAnd:
        return m.LogicalAnd(*[self.rec(child)
                              for child in expr.children])

    def map_comparison(self, expr: p.Comparison) -> m.Comparison:
        # pylint: disable=too-many-function-args
        return m.Comparison(self.rec(expr.left),
                            m.ComparisonOp(expr.operator),
                            self.rec(expr.right),
                            )

    def map_if(self, expr: p.If) -> m.If:
        # pylint: disable=too-many-function-args
        return m.If(self.rec(expr.condition),
                    self.rec(expr.then),
                    self.rec(expr.else_))

    def map_dot_wildcard(self, expr: p.DotWildcard) -> m.Wildcard:
        return m.Wildcard.dot(expr.name)

    def map_star_wildcard(self, expr: p.StarWildcard) -> m.Wildcard:
        return m.Wildcard.star(expr.name)

# }}}


# {{{ from matchpy

class FromMatchpyExpressionMapper(BaseMatchPyMapper):
    def map_scalar(self, expr: m.Scalar) -> PbScalar:
        return expr.value

    def map_variable(self, expr: m.Variable) -> p.Variable:
        return p.Variable(expr.id.value)

    def map_call(self, expr: m.Call) -> p.Call:
        return p.Call(self.rec(expr.function),
                      tuple(self.rec(arg)
                            for arg in expr.args._operands))

    def map_subscript(self, expr: m.Subscript) -> p.Subscript:
        return p.Subscript(self.rec(expr.aggregate),
                           tuple(self.rec(idx)
                                 for idx in expr.indices))

    def map_true_div(self, expr: m.TrueDiv) -> p.Quotient:
        return p.Quotient(self.rec(expr.x1), self.rec(expr.x2))

    def map_floor_div(self, expr: m.FloorDiv) -> p.FloorDiv:
        return p.FloorDiv(self.rec(expr.x1), self.rec(expr.x2))

    def map_modulo(self, expr: m.Modulo) -> p.Remainder:
        return p.Remainder(self.rec(expr.x1), self.rec(expr.x2))

    def map_power(self, expr: m.Power) -> p.Power:
        return p.Power(self.rec(expr.x1), self.rec(expr.x2))

    def map_left_shift(self, expr: m.LeftShift) -> p.LeftShift:
        return p.LeftShift(self.rec(expr.x1), self.rec(expr.x2))

    def map_right_shift(self, expr: m.RightShift) -> p.RightShift:
        return p.RightShift(self.rec(expr.x1), self.rec(expr.x2))

    def map_sum(self, expr: m.Sum) -> p.Sum:
        return p.Sum(tuple(self.rec(child)
                           for child in expr.children))

    def map_product(self, expr: m.Product) -> p.Product:
        return p.Product(tuple(self.rec(operand)
                               for operand in expr.operands))

    def map_logical_or(self, expr: m.LogicalOr) -> p.LogicalOr:
        return p.LogicalOr(tuple(self.rec(operand)
                                 for operand in expr.operands))

    def map_logical_and(self, expr: m.LogicalAnd) -> p.LogicalAnd:
        return p.LogicalAnd(tuple(self.rec(operand)
                                  for operand in expr.operands))

    def map_bitwise_or(self, expr: m.BitwiseOr) -> p.BitwiseOr:
        return p.BitwiseOr(tuple(self.rec(operand)
                                 for operand in expr.operands))

    def map_bitwise_and(self, expr: m.BitwiseAnd) -> p.BitwiseAnd:
        return p.BitwiseAnd(tuple(self.rec(operand)
                                  for operand in expr.operands))

    def map_bitwise_xor(self, expr: m.BitwiseXor) -> p.BitwiseXor:
        return p.BitwiseXor(tuple(self.rec(operand)
                                  for operand in expr.operands))

    def map_logical_not(self, expr: m.LogicalNot) -> p.LogicalNot:
        return p.LogicalNot(self.rec(expr.x))

    def map_bitwise_not(self, expr: m.BitwiseNot) -> p.BitwiseNot:
        return p.BitwiseNot(self.rec(expr.x))

    def map_comparison(self, expr: m.Comparison) -> p.Comparison:
        return p.Comparison(self.rec(expr.left),
                            expr.operator.value,
                            self.rec(expr.right))

    def map_if(self, expr: m.If) -> p.If:
        return p.If(self.rec(expr.condition),
                    self.rec(expr.then),
                    self.rec(expr.else_))

# }}}


@dataclass(frozen=True, eq=True)
class ToFromReplacement:
    f: Callable[..., p.ExpressionNode]
    to_matchpy_expr: m.ToMatchpyT
    from_matchpy_expr: m.FromMatchpyT

    def __call__(self, **kwargs):
        kwargs_to_f = {}

        for kw, arg in kwargs.items():
            if isinstance(arg, MatchpyExpression):
                arg = self.from_matchpy_expr(arg)
            elif isinstance(arg, multiset.Multiset):
                arg = multiset.Multiset({self.from_matchpy_expr(expr): count
                                         for expr, count in arg.items()})
            elif isinstance(arg, tuple):
                arg = tuple(self.from_matchpy_expr(el) for el in arg)
            else:
                raise NotImplementedError(f"Cannot convert back {type(arg)}")

            kwargs_to_f[kw] = arg

        return self.to_matchpy_expr(self.f(**kwargs_to_f))

# vim: foldmethod=marker



"""
.. autoclass:: SubstitutionMapper
.. autoclass:: CachedSubstitutionMapper
.. autofunction:: make_subst_func
.. autofunction:: substitute

.. autoclass:: Callable[[AlgebraicLeaf], Expression | None]

References
----------

.. class:: SupportsGetItem

    A protocol with a ``__getitem__`` method.
"""

from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from pymbolic.mapper import CachedIdentityMapper, IdentityMapper
from pymbolic.primitives import AlgebraicLeaf
from pymbolic.typing import Expression


if TYPE_CHECKING:
    from useful_types import SupportsGetItem, SupportsItems


class SubstitutionMapper(IdentityMapper[[]]):
    def __init__(
        self, subst_func: Callable[[AlgebraicLeaf], Expression | None]
    ) -> None:
        self.subst_func = subst_func

    def map_variable(self, expr):
        result = self.subst_func(expr)
        if result is not None:
            return result
        else:
            return expr

    def map_subscript(self, expr):
        result = self.subst_func(expr)
        if result is not None:
            return result
        else:
            return IdentityMapper.map_subscript(self, expr)

    def map_lookup(self, expr):
        result = self.subst_func(expr)
        if result is not None:
            return result
        else:
            return IdentityMapper.map_lookup(self, expr)


class CachedSubstitutionMapper(CachedIdentityMapper[[]], SubstitutionMapper):
    def __init__(
        self, subst_func: Callable[[AlgebraicLeaf], Expression | None]
    ) -> None:
        # FIXME Mypy says:
        # error: Argument 1 to "__init__" of "CachedMapper" has incompatible type
        # "CachedSubstitutionMapper"; expected "CachedMapper[ResultT, P]"  [arg-type]
        # This seems spurious?
        CachedIdentityMapper.__init__(self)  # type: ignore[arg-type]
        SubstitutionMapper.__init__(self, subst_func)


def make_subst_func(
    # "Any" here avoids the whole Mapping variance disaster
    # e.g. https://github.com/python/typing/issues/445
    variable_assignments: SupportsGetItem[Any, Expression],
) -> Callable[[AlgebraicLeaf], Expression | None]:
    import pymbolic.primitives as primitives

    def subst_func(var: AlgebraicLeaf) -> Expression | None:
        try:
            return variable_assignments[var]
        except KeyError:
            if isinstance(var, primitives.Variable):
                try:
                    return variable_assignments[var.name]
                except KeyError:
                    return None
            else:
                return None

    return subst_func


def substitute(
    expression: Expression,
    variable_assignments: SupportsItems[AlgebraicLeaf | str, Expression] | None
    = None,
    mapper_cls=CachedSubstitutionMapper,
    **kwargs: Expression,
):
    """
    :arg mapper_cls: A :class:`type` of the substitution mapper
        whose instance applies the substitution.
    """
    if variable_assignments is None:
        # "Any" here avoids pointless grief about variance
        # e.g. https://github.com/python/typing/issues/445
        v_ass_copied: dict[Any, Expression] = {}
    else:
        v_ass_copied = dict(variable_assignments.items())

    v_ass_copied.update(kwargs)

    return mapper_cls(make_subst_func(v_ass_copied))(expression)



from __future__ import annotations


__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable, Iterable, Mapping, Set
from typing import (
    TYPE_CHECKING,
    Concatenate,
    Generic,
    TypeAlias,
    TypeVar,
    cast,
)
from warnings import warn

from immutabledict import immutabledict
from typing_extensions import ParamSpec, TypeIs

import pymbolic.primitives as p
from pymbolic.typing import ArithmeticExpression, Expression


if TYPE_CHECKING:
    import numpy as np

    from pymbolic.geometric_algebra import MultiVector
    from pymbolic.rational import Rational


__doc__ = """
Basic dispatch
--------------

.. class:: ResultT

    A type variable for the result returned by a :class:`Mapper`.

.. autoclass:: Mapper

    .. automethod:: __call__

    .. method:: rec(expr, *args, **kwargs)

        Identical to :meth:`__call__`, but intended for use in recursive dispatch
        in mapper methods.

    .. automethod:: handle_unsupported_expression

    .. rubric:: Handling objects that don't declare mapper methods

    In particular, this includes many non-subclasses of
    :class:`pymbolic.ExpressionNode`.

    .. automethod:: map_foreign

    These are abstract methods for foreign objects that should be overridden
    in subclasses:

    .. method:: map_constant(expr, *args, **kwargs)

        Mapper method for constants.
        See :func:`pymbolic.primitives.register_constant_class`.

    .. method:: map_list(expr, *args, **kwargs)

    .. method:: map_tuple(expr, *args, **kwargs)

    .. method:: map_numpy_array(expr, *args, **kwargs)

Base classes for new mappers
----------------------------

.. autoclass:: CombineMapper

.. autoclass:: Collector

.. autoclass:: IdentityMapper

.. autoclass:: WalkMapper

.. autoclass:: CSECachingMapperMixin


Base classes for mappers with memoization support
-------------------------------------------------

.. autoclass:: CachedMapper

.. autoclass:: CachedIdentityMapper

.. autoclass:: CachedCombineMapper

.. autoclass:: CachedCollector

.. autoclass:: CachedWalkMapper
"""


if TYPE_CHECKING:
    import numpy as np

    def is_numpy_array(val) -> TypeIs[np.ndarray]:
        return isinstance(val, np.ndarray)
else:
    try:
        import numpy as np

        def is_numpy_array(val):
            return isinstance(val, np.ndarray)
    except ImportError:
        def is_numpy_array(ary):
            return False


class UnsupportedExpressionError(ValueError):
    pass


# {{{ mapper base

ResultT = TypeVar("ResultT")

# This ParamSpec could be marked contravariant (just like Callable is contravariant
# in its arguments). As of mypy 1.14/Py3.13 (Nov 2024), mypy complains of as-yet
# undefined semantics, so it's probably too soon.
P = ParamSpec("P")


class Mapper(Generic[ResultT, P]):
    """A visitor for trees of :class:`pymbolic.ExpressionNode`
    subclasses. Each expression-derived object is dispatched to the
    method named by the :attr:`pymbolic.ExpressionNode.mapper_method`
    attribute and if not found, the methods named by the class attribute
    *mapper_method* in the method resolution order of the object.

    ..automethod:: handle_unsupported_expression
    ..automethod:: __call__
    ..automethod:: rec
    """

    def handle_unsupported_expression(self,
            expr: object, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        """Mapper method that is invoked for
        :class:`pymbolic.ExpressionNode` subclasses for which a mapper
        method does not exist in this mapper.
        """

        raise UnsupportedExpressionError(
                "{} cannot handle expressions of type {}".format(
                    type(self), type(expr)))

    def __call__(self,
             expr: Expression, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        """Dispatch *expr* to its corresponding mapper method. Pass on
        ``*args`` and ``**kwargs`` unmodified.

        This method is intended as the top-level dispatch entry point and may
        be overridden by subclasses to present a different/more convenient
        interface. :meth:`rec` on the other hand is intended as the recursive
        dispatch method to be used to recurse within mapper method
        implementations.
        """

        method_name = getattr(expr, "mapper_method", None)
        if method_name is not None:
            method = getattr(self, method_name, None)
            if method is not None:
                result = method(expr, *args, **kwargs)
                return result

        if isinstance(expr, p.ExpressionNode):
            for cls in type(expr).__mro__[1:]:
                method_name = getattr(cls, "mapper_method", None)
                if method_name:
                    method = getattr(self, method_name, None)
                    if method:
                        return method(expr, *args, **kwargs)
            else:
                return self.handle_unsupported_expression(expr, *args, **kwargs)
        else:
            return self.map_foreign(expr, *args, **kwargs)

    rec = __call__

    def rec_fallback(self,
            expr: object, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        if isinstance(expr, p.ExpressionNode):
            for cls in type(expr).__mro__[1:]:
                method_name = getattr(cls, "mapper_method", None)
                if method_name:
                    method = getattr(self, method_name, None)
                    if method:
                        return method(expr, *args, **kwargs)
            else:
                return self.handle_unsupported_expression(expr, *args, **kwargs)
        else:
            return self.map_foreign(expr, *args, **kwargs)

    def map_algebraic_leaf(self,
            expr: p.AlgebraicLeaf,
            *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_variable(self,
            expr: p.Variable, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.map_algebraic_leaf(expr, *args, **kwargs)

    def map_subscript(self,
            expr: p.Subscript, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.map_algebraic_leaf(expr, *args, **kwargs)

    def map_call(self,
            expr: p.Call, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.map_algebraic_leaf(expr, *args, **kwargs)

    def map_call_with_kwargs(self,
            expr: p.CallWithKwargs,
            *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.map_algebraic_leaf(expr, *args, **kwargs)

    def map_lookup(self,
            expr: p.Lookup, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.map_algebraic_leaf(expr, *args, **kwargs)

    def map_if(self,
            expr: p.If, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_sum(self,
            expr: p.Sum, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_product(self,
            expr: p.Product, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_rational(self,
            expr: Rational, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_quotient(self,
            expr: p.Quotient, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_floor_div(self,
            expr: p.FloorDiv, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_remainder(self,
            expr: p.Remainder, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_constant(self,
            expr: object, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_comparison(self,
            expr: p.Comparison, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_min(self,
            expr: p.Min, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_max(self,
            expr: p.Max, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_list(self,
            expr: list[Expression], *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_tuple(self,
            expr: tuple[Expression, ...],
            *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_numpy_array(self,
            expr: np.ndarray, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        raise NotImplementedError

    def map_left_shift(self,
            expr: p.LeftShift, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        raise NotImplementedError

    def map_right_shift(self,
                expr: p.RightShift, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        raise NotImplementedError

    def map_bitwise_not(self,
                expr: p.BitwiseNot, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        raise NotImplementedError

    def map_bitwise_or(self,
                expr: p.BitwiseOr, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        raise NotImplementedError

    def map_bitwise_and(self,
                expr: p.BitwiseAnd, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        raise NotImplementedError

    def map_bitwise_xor(self,
                expr: p.BitwiseXor, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        raise NotImplementedError

    def map_logical_not(self,
                expr: p.LogicalNot, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        raise NotImplementedError

    def map_logical_or(self,
                expr: p.LogicalOr, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        raise NotImplementedError

    def map_logical_and(self,
                expr: p.LogicalAnd, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        raise NotImplementedError

    def map_nan(self,
                expr: p.NaN,
                *args: P.args,
                **kwargs: P.kwargs
            ) -> ResultT:
        return self.map_algebraic_leaf(expr, *args, **kwargs)

    def map_foreign(self,
                expr: object,
                *args: P.args,
                **kwargs: P.kwargs
            ) -> ResultT:
        """Mapper method dispatch for non-:mod:`pymbolic` objects."""

        if isinstance(expr, p.VALID_CONSTANT_CLASSES):
            return self.map_constant(expr, *args, **kwargs)
        elif is_numpy_array(expr):
            return self.map_numpy_array(expr, *args, **kwargs)
        elif isinstance(expr, tuple):
            return self.map_tuple(expr, *args, **kwargs)
        elif isinstance(expr, list):
            warn("List found in expression graph. "
                 "This is deprecated and will stop working in 2025. "
                 "Use tuples instead.", DeprecationWarning, stacklevel=2
             )
            return self.map_list(expr, *args, **kwargs)
        else:
            raise ValueError(
                    "{} encountered invalid foreign object: {}".format(
                        self.__class__, repr(expr)))


class _NotInCache:
    pass


CacheKeyT: TypeAlias = Hashable


class CachedMapper(Mapper[ResultT, P]):
    """
    A mapper that memoizes the mapped result for the expressions traversed.

    .. automethod:: get_cache_key
    """
    def __init__(self) -> None:
        self._cache: dict[CacheKeyT, ResultT] = {}
        Mapper.__init__(self)

    def get_cache_key(self,
              expr: Expression,
              *args: P.args,
              **kwargs: P.kwargs
          ) -> CacheKeyT:
        """
        Returns the key corresponding to which the result of a mapper method is
        stored in the cache.

        .. warning::

            Assumes that elements of *args* and *kwargs* are immutable, and that
            *self* does not store any mutable state. Derived mappers must
            override this method.
        """
        # Must add 'type(expr)', to differentiate between python scalar types.
        # In Python, the following conditions are true: "hash(4) == hash(4.0)"
        # and "4 == 4.0", but their traversal results cannot be re-used.
        return (type(expr), expr, args, immutabledict(kwargs))

    def __call__(self,
                 expr: Expression,
                 *args: P.args,
                 **kwargs: P.kwargs
             ) -> ResultT:
        result = self._cache.get(
                (cache_key := self.get_cache_key(expr, *args, **kwargs)),
                _NotInCache)
        if not isinstance(result, type):
            return result

        method_name = getattr(expr, "mapper_method", None)
        if method_name is not None:
            method = cast(
                Callable[Concatenate[Expression, P], ResultT] | None,
                getattr(self, method_name, None)
                )
            if method is not None:
                result = method(expr, *args, **kwargs)
                self._cache[cache_key] = result
                return result

        result = self.rec_fallback(expr, *args, **kwargs)
        self._cache[cache_key] = result
        return result

    rec = __call__

# }}}


# {{{ combine mapper

class CombineMapper(Mapper[ResultT, P]):
    """A mapper whose goal it is to *combine* all branches of the expression
    tree into one final result. The default implementation of all mapper
    methods simply recurse (:meth:`Mapper.rec`) on all branches emanating from
    the current expression, and then call :meth:`combine` on a tuple of
    results.

    .. method:: combine(values)

        Combine the mapped results of multiple expressions (given in *values*)
        into a single result, often by summing or taking set unions.

    The :class:`pymbolic.mapper.flop_counter.FlopCounter` is a very simple
    example.  (Look at its source for an idea of how to derive from
    :class:`CombineMapper`.) The
    :class:`pymbolic.mapper.dependency.DependencyMapper` is another example.
    """

    def combine(self, values: Iterable[ResultT]) -> ResultT:
        raise NotImplementedError

    def map_call(self,
            expr: p.Call, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine((
            self.rec(expr.function, *args, **kwargs),
            *[self.rec(child, *args, **kwargs) for child in expr.parameters]
            ))

    def map_call_with_kwargs(self,
            expr: p.CallWithKwargs,
            *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine((
            self.rec(expr.function, *args, **kwargs),
            *[self.rec(child, *args, **kwargs) for child in expr.parameters],
            *[self.rec(child, *args, **kwargs)
              for child in expr.kw_parameters.values()]
            ))

    def map_subscript(self,
            expr: p.Subscript, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine(
                [self.rec(expr.aggregate, *args, **kwargs),
                    self.rec(expr.index, *args, **kwargs)])

    def map_lookup(self,
            expr: p.Lookup, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.rec(expr.aggregate, *args, **kwargs)

    def map_sum(self,
            expr: p.Sum, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine(self.rec(child, *args, **kwargs)
                for child in expr.children)

    def map_product(self,
            expr: p.Product, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine(self.rec(child, *args, **kwargs)
                for child in expr.children)

    def map_quotient(self,
            expr: p.Quotient, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine((
            self.rec(expr.numerator, *args, **kwargs),
            self.rec(expr.denominator, *args, **kwargs)))

    def map_floor_div(self,
            expr: p.FloorDiv, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine((
            self.rec(expr.numerator, *args, **kwargs),
            self.rec(expr.denominator, *args, **kwargs)))

    def map_remainder(self,
            expr: p.Remainder, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine((
            self.rec(expr.numerator, *args, **kwargs),
            self.rec(expr.denominator, *args, **kwargs)))

    def map_power(self,
            expr: p.Power, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine((
                self.rec(expr.base, *args, **kwargs),
                self.rec(expr.exponent, *args, **kwargs)))

    def map_left_shift(self,
            expr: p.LeftShift, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine((
            self.rec(expr.shiftee, *args, **kwargs),
            self.rec(expr.shift, *args, **kwargs)))

    def map_right_shift(self,
            expr: p.RightShift, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine((
            self.rec(expr.shiftee, *args, **kwargs),
            self.rec(expr.shift, *args, **kwargs)))

    def map_bitwise_not(self,
            expr: p.BitwiseNot, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.rec(expr.child, *args, **kwargs)

    def map_bitwise_or(self,
            expr: p.BitwiseOr, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine(self.rec(child, *args, **kwargs)
                for child in expr.children)

    def map_bitwise_and(self,
            expr: p.BitwiseAnd, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine(self.rec(child, *args, **kwargs)
                for child in expr.children)

    def map_bitwise_xor(self,
            expr: p.BitwiseXor, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine(self.rec(child, *args, **kwargs)
                for child in expr.children)

    def map_logical_not(self,
            expr: p.LogicalNot, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.rec(expr.child, *args, **kwargs)

    def map_logical_or(self,
            expr: p.LogicalOr, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine(self.rec(child, *args, **kwargs)
                for child in expr.children)

    def map_logical_and(self,
            expr: p.LogicalAnd, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine(self.rec(child, *args, **kwargs)
                for child in expr.children)

    def map_comparison(self,
            expr: p.Comparison, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine((
            self.rec(expr.left, *args, **kwargs),
            self.rec(expr.right, *args, **kwargs)))

    def map_max(self,
            expr: p.Max, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine(self.rec(child, *args, **kwargs)
                for child in expr.children)

    def map_min(self,
            expr: p.Min, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine(self.rec(child, *args, **kwargs)
                for child in expr.children)

    def map_tuple(self,
                expr: tuple[Expression, ...], *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        return self.combine(self.rec(child, *args, **kwargs) for child in expr)

    def map_list(self,
                expr: list[Expression], *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        return self.combine(self.rec(child, *args, **kwargs) for child in expr)

    def map_numpy_array(self,
                expr: np.ndarray, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        return self.combine(self.rec(el, *args, **kwargs) for el in expr.flat)

    def map_multivector(self,
                expr: MultiVector[ArithmeticExpression],
                *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        return self.combine(
                self.rec(coeff, *args, **kwargs)
                for bits, coeff in expr.data.items())

    def map_common_subexpression(self,
                expr: p.CommonSubexpression, *args: P.args, **kwargs: P.kwargs
            ) -> ResultT:
        return self.rec(expr.child, *args, **kwargs)

    def map_if(self,
            expr: p.If, *args: P.args, **kwargs: P.kwargs) -> ResultT:
        return self.combine([
            self.rec(expr.condition, *args, **kwargs),
            self.rec(expr.then, *args, **kwargs),
            self.rec(expr.else_, *args, **kwargs)])


class CachedCombineMapper(CachedMapper, CombineMapper):
    pass

# }}}


# {{{ collector

CollectedT = TypeVar("CollectedT")


class Collector(CombineMapper[Set[CollectedT], P]):
    """A subclass of :class:`CombineMapper` for the common purpose of
    collecting data derived from an expression in a set that gets 'unioned'
    across children at each non-leaf node in the expression tree.

    By default, nothing is collected. All leaves return empty sets.

    .. versionadded:: 2014.3
    """

    def combine(self,
                values: Iterable[Set[CollectedT]]
            ) -> Set[CollectedT]:
        import operator
        from functools import reduce
        return reduce(operator.or_, values, set())

    def map_constant(self, expr: object,
                     *args: P.args, **kwargs: P.kwargs) -> Set[CollectedT]:
        return set()

    def map_variable(self, expr: p.Variable,
                     *args: P.args, **kwargs: P.kwargs) -> Set[CollectedT]:
        return set()

    def map_wildcard(self, expr: p.Wildcard,
                     *args: P.args, **kwargs: P.kwargs) -> Set[CollectedT]:
        return set()

    def map_dot_wildcard(self, expr: p.DotWildcard,
                     *args: P.args, **kwargs: P.kwargs) -> Set[CollectedT]:
        return set()

    def map_star_wildcard(self, expr: p.StarWildcard,
                     *args: P.args, **kwargs: P.kwargs) -> Set[CollectedT]:
        return set()

    def map_function_symbol(self, expr: p.FunctionSymbol,
                     *args: P.args, **kwargs: P.kwargs) -> Set[CollectedT]:
        return set()


class CachedCollector(CachedMapper, Collector):
    pass

# }}}


# {{{ identity mapper

class IdentityMapper(Mapper[Expression, P]):
    """A :class:`Mapper` whose default mapper methods
    make a deep copy of each subexpression.

    See :ref:`custom-manipulation` for an example of the
    manipulations that can be implemented this way.

    .. automethod:: rec_arith
    """

    def rec_arith(self,
                expr: ArithmeticExpression, *args: P.args, **kwargs: P.kwargs
            ) -> ArithmeticExpression:
        res = self.rec(expr, *args, **kwargs)
        assert p.is_arithmetic_expression(res)
        return res

    def map_constant(self,
                expr: object, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        # leaf -- no need to rebuild
        assert p.is_valid_operand(expr)
        return expr

    def map_variable(self,
                expr: p.Variable, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        # leaf -- no need to rebuild
        return expr

    def map_wildcard(self,
                expr: p.Wildcard, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        return expr

    def map_dot_wildcard(self,
                expr: p.DotWildcard, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        return expr

    def map_star_wildcard(self,
                expr: p.StarWildcard, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        return expr

    def map_function_symbol(self,
                expr: p.FunctionSymbol, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        return expr

    def map_call(self,
                expr: p.Call, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        function = self.rec(expr.function, *args, **kwargs)
        parameters = tuple([
            self.rec(child, *args, **kwargs) for child in expr.parameters
            ])
        if (function is expr.function
            and all(child is orig_child
                for child, orig_child in zip(
                            expr.parameters, parameters, strict=True))):
            return expr

        return type(expr)(function, parameters)

    def map_call_with_kwargs(self,
                expr: p.CallWithKwargs, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        function = self.rec(expr.function, *args, **kwargs)
        parameters = tuple([
            self.rec(child, *args, **kwargs) for child in expr.parameters
            ])
        kw_parameters: Mapping[str, Expression] = immutabledict({
                key: self.rec(val, *args, **kwargs)
                for key, val in expr.kw_parameters.items()})

        if (function is expr.function
            and all(child is orig_child for child, orig_child in
                zip(parameters, expr.parameters, strict=True))
                and all(kw_parameters[k] is v for k, v in
                        expr.kw_parameters.items())):
            return expr
        return type(expr)(function, parameters, kw_parameters)

    def map_subscript(self,
                expr: p.Subscript, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        aggregate = self.rec(expr.aggregate, *args, **kwargs)
        index = self.rec(expr.index, *args, **kwargs)
        if aggregate is expr.aggregate and index is expr.index:
            return expr
        return type(expr)(aggregate, index)

    def map_lookup(self,
                expr: p.Lookup, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        aggregate = self.rec(expr.aggregate, *args, **kwargs)
        if aggregate is expr.aggregate:
            return expr
        return type(expr)(aggregate, expr.name)

    def map_sum(self,
                expr: p.Sum, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        children = [self.rec(child, *args, **kwargs) for child in expr.children]
        if all(child is orig_child
                for child, orig_child in zip(children, expr.children, strict=True)):
            return expr

        return type(expr)(tuple(children))

    def map_product(self,
                expr: p.Product, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        children = [self.rec(child, *args, **kwargs) for child in expr.children]
        if all(child is orig_child
                for child, orig_child in zip(children, expr.children, strict=True)):
            return expr

        return type(expr)(tuple(children))

    def map_quotient(self,
                expr: p.Quotient, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        numerator = self.rec_arith(expr.numerator, *args, **kwargs)
        denominator = self.rec_arith(expr.denominator, *args, **kwargs)
        if numerator is expr.numerator and denominator is expr.denominator:
            return expr
        return expr.__class__(numerator, denominator)

    def map_floor_div(self,
                expr: p.FloorDiv, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        numerator = self.rec_arith(expr.numerator, *args, **kwargs)
        denominator = self.rec_arith(expr.denominator, *args, **kwargs)
        if numerator is expr.numerator and denominator is expr.denominator:
            return expr
        return expr.__class__(numerator, denominator)

    def map_remainder(self,
                expr: p.Remainder, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        numerator = self.rec_arith(expr.numerator, *args, **kwargs)
        denominator = self.rec_arith(expr.denominator, *args, **kwargs)
        if numerator is expr.numerator and denominator is expr.denominator:
            return expr
        return expr.__class__(numerator, denominator)

    def map_power(self,
                expr: p.Power, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        base = self.rec_arith(expr.base, *args, **kwargs)
        exponent = self.rec_arith(expr.exponent, *args, **kwargs)
        if base is expr.base and exponent is expr.exponent:
            return expr
        return expr.__class__(base, exponent)

    def map_left_shift(self,
                expr: p.LeftShift, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        shiftee = self.rec(expr.shiftee, *args, **kwargs)
        shift = self.rec(expr.shift, *args, **kwargs)
        if shiftee is expr.shiftee and shift is expr.shift:
            return expr
        return type(expr)(shiftee, shift)

    def map_right_shift(self,
                expr: p.RightShift, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        shiftee = self.rec(expr.shiftee, *args, **kwargs)
        shift = self.rec(expr.shift, *args, **kwargs)
        if shiftee is expr.shiftee and shift is expr.shift:
            return expr
        return type(expr)(shiftee, shift)

    def map_bitwise_not(self,
                expr: p.BitwiseNot, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        child = self.rec(expr.child, *args, **kwargs)
        if child is expr.child:
            return expr
        return type(expr)(child)

    def map_bitwise_or(self,
                expr: p.BitwiseOr, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        children = [self.rec(child, *args, **kwargs) for child in expr.children]
        if all(child is orig_child
                for child, orig_child in zip(children, expr.children, strict=True)):
            return expr

        return type(expr)(tuple(children))

    def map_bitwise_and(self,
                expr: p.BitwiseAnd, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        children = [self.rec(child, *args, **kwargs) for child in expr.children]
        if all(child is orig_child
                for child, orig_child in zip(children, expr.children, strict=True)):
            return expr

        return type(expr)(tuple(children))

    def map_bitwise_xor(self,
                expr: p.BitwiseXor, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        children = [self.rec(child, *args, **kwargs) for child in expr.children]
        if all(child is orig_child
                for child, orig_child in zip(children, expr.children, strict=True)):
            return expr

        return type(expr)(tuple(children))

    def map_logical_not(self,
                expr: p.LogicalNot, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        child = self.rec(expr.child, *args, **kwargs)
        if child is expr.child:
            return expr
        return type(expr)(child)

    def map_logical_or(self,
                expr: p.LogicalOr, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        children = [self.rec(child, *args, **kwargs) for child in expr.children]
        if all(child is orig_child
                for child, orig_child in zip(children, expr.children, strict=True)):
            return expr

        return type(expr)(tuple(children))

    def map_logical_and(self,
                expr: p.LogicalAnd, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        children = [self.rec(child, *args, **kwargs) for child in expr.children]
        if all(child is orig_child
                for child, orig_child in zip(children, expr.children, strict=True)):
            return expr

        return type(expr)(tuple(children))

    def map_comparison(self,
                expr: p.Comparison, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        left = self.rec(expr.left, *args, **kwargs)
        right = self.rec(expr.right, *args, **kwargs)
        if left is expr.left and right is expr.right:
            return expr

        return type(expr)(left, expr.operator, right)

    def map_list(self,
                expr: list[Expression], *args: P.args, **kwargs: P.kwargs
            ) -> Expression:

        # True fact: lists aren't expressions
        return [self.rec(child, *args, **kwargs) for child in expr]  # type: ignore[return-value]

    def map_tuple(self,
                expr: tuple[Expression, ...], *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        children = [self.rec(child, *args, **kwargs) for child in expr]
        if all(child is orig_child
                for child, orig_child in zip(children, expr, strict=True)):
            return expr

        return tuple(children)

    def map_numpy_array(self,
                expr: np.ndarray, *args: P.args, **kwargs: P.kwargs
            ) -> Expression:

        import numpy
        result = numpy.empty(expr.shape, dtype=object)
        for i in numpy.ndindex(expr.shape):
            result[i] = self.rec(expr[i], *args, **kwargs)

        # True fact: ndarrays aren't expressions
        return result  # type: ignore[return-value]

    def map_multivector(self,
                expr: MultiVector[ArithmeticExpression],
                *args: P.args, **kwargs: P.kwargs
            ) -> Expression:
        # True fact: MultiVectors aren't expressions
        return expr.map(lambda ch: cast(ArithmeticExpression,
                                        self.rec(ch, *args, **kwargs)))  # type: ignore[return-value]

    def map_common_subexpression(self,
                expr: p.CommonSubexpression,
                *args: P.args, **kwargs: P.kwargs) -> Expression:
        result = self.rec(expr.child, *args, **kwargs)
        if result is expr.child:
            return expr

        return type(expr)(
                result,
                expr.prefix,
                expr.scope,
                **expr.get_extra_properties())

    def map_substitution(self,
                 expr: p.Substitution,
                 *args: P.args, **kwargs: P.kwargs) -> Expression:
        child = self.rec(expr.child, *args, **kwargs)
        values = tuple([self.rec(v, *args, **kwargs) for v in expr.values])
        if child is expr.child and all(val is orig_val
                for val, orig_val in zip(values, expr.values, strict=True)):
            return expr

        return type(expr)(child, expr.variables, values)

    def map_derivative(self,
                expr: p.Derivative,
                *args: P.args, **kwargs: P.kwargs) -> Expression:
        child = self.rec(expr.child, *args, **kwargs)
        if child is expr.child:
            return expr

        return type(expr)(child, expr.variables)

    def map_slice(self,
                expr: p.Slice,
                *args: P.args, **kwargs: P.kwargs) -> Expression:
        children: p.SliceChildrenT = cast(p.SliceChildrenT, tuple([
            None if child is None else self.rec(child, *args, **kwargs)
            for child in expr.children
            ]))
        if all(child is orig_child
                for child, orig_child in zip(children, expr.children, strict=True)):
            return expr

        return type(expr)(children)

    def map_if(self, expr: p.If, *args: P.args, **kwargs: P.kwargs) -> Expression:
        condition = self.rec(expr.condition, *args, **kwargs)
        then = self.rec(expr.then, *args, **kwargs)
        else_ = self.rec(expr.else_, *args, **kwargs)
        if condition is expr.condition \
                and then is expr.then \
                and else_ is expr.else_:
            return expr

        return type(expr)(condition, then, else_)

    def map_min(self,
                expr: p.Min, *args: P.args, **kwargs: P.kwargs) -> Expression:
        children = tuple([
            self.rec(child, *args, **kwargs) for child in expr.children
            ])
        if all(child is orig_child
                for child, orig_child in zip(children, expr.children, strict=True)):
            return expr

        return type(expr)(children)

    def map_max(self,
                expr: p.Max, *args: P.args, **kwargs: P.kwargs) -> Expression:
        children = tuple([
            self.rec(child, *args, **kwargs) for child in expr.children
            ])
        if all(child is orig_child
                for child, orig_child in zip(children, expr.children, strict=True)):
            return expr

        return type(expr)(children)

    def map_nan(self,
                expr: p.NaN, *args: P.args, **kwargs: P.kwargs) -> Expression:
        # Leaf node -- don't recurse
        return expr


class CachedIdentityMapper(CachedMapper[Expression, P], IdentityMapper[P]):
    pass

# }}}


# {{{ walk mapper

class WalkMapper(Mapper[None, P]):
    """A mapper whose default mapper method implementations simply recurse
    without propagating any result. Also calls :meth:`visit` for each
    visited subexpression.

    ``map_...`` methods are required to call :meth:`visit` *before*
        descending to visit their children.

    .. method:: visit(expr, *args, **kwargs)

        Returns *False* if no children of this node should be examined.

    .. method:: post_visit(expr, *args, **kwargs)

        Is called after a node's children are visited.
    """

    def map_constant(self, expr: object, *args: P.args, **kwargs: P.kwargs) -> None:
        self.visit(expr, *args, **kwargs)
        self.post_visit(expr, *args, **kwargs)

    def map_variable(self, expr: p.Variable, *args: P.args, **kwargs: P.kwargs) -> None:
        self.visit(expr, *args, **kwargs)
        self.post_visit(expr, *args, **kwargs)

    def map_wildcard(self, expr: p.Wildcard, *args: P.args, **kwargs: P.kwargs) -> None:
        self.visit(expr, *args, **kwargs)
        self.post_visit(expr, *args, **kwargs)

    def map_dot_wildcard(self,
            expr: p.DotWildcard, *args: P.args, **kwargs: P.kwargs) -> None:
        self.visit(expr, *args, **kwargs)
        self.post_visit(expr, *args, **kwargs)

    def map_star_wildcard(self,
            expr: p.StarWildcard, *args: P.args, **kwargs: P.kwargs) -> None:
        self.visit(expr, *args, **kwargs)
        self.post_visit(expr, *args, **kwargs)

    def map_function_symbol(self,
            expr: p.FunctionSymbol, *args: P.args, **kwargs: P.kwargs) -> None:
        self.visit(expr, *args, **kwargs)
        self.post_visit(expr, *args, **kwargs)

    def map_nan(self,
            expr: p.NaN, *args: P.args, **kwargs: P.kwargs) -> None:
        self.visit(expr, *args, **kwargs)
        self.post_visit(expr, *args, **kwargs)

    def map_call(self, expr: p.Call, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.function, *args, **kwargs)
        for child in expr.parameters:
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_call_with_kwargs(self,
                expr: p.CallWithKwargs,
                *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.function, *args, **kwargs)
        for child in expr.parameters:
            self.rec(child, *args, **kwargs)

        for child in list(expr.kw_parameters.values()):
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_subscript(self,
                expr: p.Subscript,
                *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.aggregate, *args, **kwargs)
        self.rec(expr.index, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_lookup(self,
                expr: p.Lookup, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.aggregate, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_sum(self, expr: p.Sum, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        for child in expr.children:
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_product(self, expr: p.Product, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        for child in expr.children:
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_quotient(self, expr: p.Quotient, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.numerator, *args, **kwargs)
        self.rec(expr.denominator, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_floor_div(self,
            expr: p.FloorDiv, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.numerator, *args, **kwargs)
        self.rec(expr.denominator, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_remainder(self,
            expr: p.Remainder, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.numerator, *args, **kwargs)
        self.rec(expr.denominator, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_power(self, expr: p.Power, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.base, *args, **kwargs)
        self.rec(expr.exponent, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_tuple(self,
                expr: tuple[Expression, ...], *args: P.args, **kwargs: P.kwargs
            ) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        for child in expr:
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_numpy_array(self,
            expr: np.ndarray, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        import numpy
        for i in numpy.ndindex(expr.shape):
            self.rec(expr[i], *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_multivector(self,
            expr: MultiVector[ArithmeticExpression],
            *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        for _bits, coeff in expr.data.items():
            self.rec(coeff, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_common_subexpression(self,
            expr: p.CommonSubexpression, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_left_shift(self,
            expr: p.LeftShift, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.shift, *args, **kwargs)
        self.rec(expr.shiftee, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_right_shift(self,
            expr: p.RightShift, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.shift, *args, **kwargs)
        self.rec(expr.shiftee, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_bitwise_not(self,
            expr: p.BitwiseNot, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_bitwise_or(self,
                expr: p.BitwiseOr, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        for child in expr.children:
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_bitwise_xor(self,
                expr: p.BitwiseXor, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        for child in expr.children:
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_bitwise_and(self,
                expr: p.BitwiseAnd, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        for child in expr.children:
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_comparison(self, expr, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.left, *args, **kwargs)
        self.rec(expr.right, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_logical_not(self,
            expr: p.LogicalNot, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_logical_or(self,
                expr: p.LogicalOr, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        for child in expr.children:
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_logical_and(self,
                expr: p.LogicalAnd, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        for child in expr.children:
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_if(self, expr, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.condition, *args, **kwargs)
        self.rec(expr.then, *args, **kwargs)
        self.rec(expr.else_, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_if_positive(self, expr, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.criterion, *args, **kwargs)
        self.rec(expr.then, *args, **kwargs)
        self.rec(expr.else_, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_min(self,
                expr: p.Min, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        for child in expr.children:
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_max(self,
                expr: p.Max, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        for child in expr.children:
            self.rec(child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_substitution(self, expr, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.child, *args, **kwargs)
        for v in expr.values:
            self.rec(v, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_derivative(self, expr, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        self.rec(expr.child, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def map_slice(self, expr, *args: P.args, **kwargs: P.kwargs) -> None:
        if not self.visit(expr, *args, **kwargs):
            return

        if expr.start is not None:
            self.rec(expr.start, *args, **kwargs)
        if expr.stop is not None:
            self.rec(expr.stop, *args, **kwargs)
        if expr.step is not None:
            self.rec(expr.step, *args, **kwargs)

        self.post_visit(expr, *args, **kwargs)

    def visit(self, expr, *args: P.args, **kwargs: P.kwargs) -> bool:
        return True

    def post_visit(self, expr, *args: P.args, **kwargs: P.kwargs) -> None:
        pass


class CachedWalkMapper(CachedMapper, WalkMapper):
    pass

# }}}


# {{{ callback mapper

# FIXME: Is it worth typing this?
class CallbackMapper(Mapper):
    def __init__(self, function, fallback_mapper):
        self.function = function
        self.fallback_mapper = fallback_mapper
        fallback_mapper.rec = self.rec

    def map_constant(self, expr, *args, **kwargs):
        return self.function(expr, self, *args, **kwargs)

    map_variable = map_constant
    map_function_symbol = map_constant
    map_call = map_constant
    map_subscript = map_constant
    map_lookup = map_constant
    map_sum = map_constant
    map_product = map_constant
    map_quotient = map_constant
    map_floor_div = map_constant
    map_remainder = map_constant
    map_power = map_constant

    map_left_shift = map_constant
    map_right_shift = map_constant

    map_bitwise_not = map_constant
    map_bitwise_or = map_constant
    map_bitwise_xor = map_constant
    map_bitwise_and = map_constant

    map_logical_not = map_constant
    map_logical_or = map_constant
    map_logical_and = map_constant

    map_list = map_constant
    map_tuple = map_constant
    map_numpy_array = map_constant
    map_common_subexpression = map_constant
    map_if_positive = map_constant
    map_if = map_constant
    map_comparison = map_constant

# }}}


# {{{ caching mixins

class CSECachingMapperMixin(ABC, Generic[ResultT, P]):
    """A :term:`mix-in` that helps
    subclassed mappers implement caching for
    :class:`pymbolic.primitives.CommonSubexpression`
    instances.

    Instead of the common mapper method for
    :class:`pymbolic.primitives.CommonSubexpression`,
    subclasses should implement the following method:

    .. method:: map_common_subexpression_uncached(expr)

    This method deliberately does not support extra arguments in mapper
    dispatch, to avoid spurious dependencies of the cache on these arguments.
    """
    _cse_cache_dict: dict[tuple[Expression, P.args, P.kwargs], ResultT]

    def map_common_subexpression(self,
                expr: p.CommonSubexpression,
                *args: P.args, **kwargs: P.kwargs) -> ResultT:
        try:
            ccd = self._cse_cache_dict
        except AttributeError:
            ccd = self._cse_cache_dict = {}

        key: tuple[Expression, P.args, P.kwargs] = (
            expr, args, immutabledict(kwargs))
        try:
            return ccd[key]
        except KeyError:
            result = self.map_common_subexpression_uncached(expr, *args, **kwargs)
            ccd[key] = result
            return result

    @abstractmethod
    def map_common_subexpression_uncached(self,
                expr: p.CommonSubexpression,
                *args: P.args, **kwargs: P.kwargs) -> ResultT:
        pass

# }}}


def __getattr__(name: str) -> object:
    if name == "RecursiveMapper":
        warn("RecursiveMapper is deprecated. Use Mapper instead. "
             "RecursiveMapper will go away in 2026.",
             DeprecationWarning, stacklevel=2)
        return Mapper

    return None

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2017 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import pymbolic.mapper


class ConstantToNumpyConversionMapper(pymbolic.mapper.IdentityMapper):
    """Because of `this numpy bug <https://github.com/numpy/numpy/issues/9438>`__,
    sized :mod:`numpy` number (i.e. ones with definite bit width, such as
    :class:`numpy.complex64`) have a low likelihood of surviving expression
    construction.

    This mapper ensures that all occurring numerical constants are of the
    expected type.
    """

    def __init__(self, real_type, complex_type=None, integer_type=None):
        import numpy as np
        self.real_type = real_type

        if complex_type is None:
            if real_type is np.float32:
                complex_type = np.complex64
            elif real_type is np.float64:
                complex_type = np.complex128
            elif real_type is np.float128:  # pylint:disable=no-member
                complex_type = np.complex256  # pylint:disable=no-member
            else:
                raise TypeError("unable to determine corresponding complex type "
                        f"for '{real_type.__name__}'")

        self.complex_type = complex_type

        self.integer_type = integer_type

    def map_constant(self, expr):
        if expr.imag:
            return self.complex_type(expr)

        expr = expr.real

        if int(expr) == expr and not isinstance(expr, float):
            if self.integer_type is not None:
                return self.integer_type(expr)
            else:
                return expr
        else:
            return self.real_type(expr)



"""Imperative program representation: transformations"""
from __future__ import annotations


__copyright__ = "Copyright (C) 2015 Matt Wala, Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


# {{{ fuse statement streams

def fuse_statement_streams_with_unique_ids(statements_a, statements_b):
    new_statements = list(statements_a)
    from pytools import UniqueNameGenerator
    stmt_id_gen = UniqueNameGenerator(
            {stmta.id for stmta in new_statements})

    b_unique_statements = []
    old_b_id_to_new_b_id = {}
    for stmtb in statements_b:
        old_id = stmtb.id
        new_id = stmt_id_gen(old_id)
        old_b_id_to_new_b_id[old_id] = new_id

        b_unique_statements.append(
                stmtb.copy(id=new_id))

    for stmtb in b_unique_statements:
        new_statements.append(
                stmtb.copy(
                    depends_on=frozenset(
                        old_b_id_to_new_b_id[dep_id]
                        for dep_id in stmtb.depends_on)))

    return new_statements, old_b_id_to_new_b_id


def fuse_instruction_streams_with_unique_ids(insns_a, insns_b):
    from warnings import warn
    warn("fuse_instruction_streams_with_unique_ids has been renamed to "
            "fuse_statement_streams_with_unique_ids", DeprecationWarning,
            stacklevel=2)

    return fuse_statement_streams_with_unique_ids(insns_a, insns_b)

# }}}


# {{{ disambiguate_identifiers

def disambiguate_identifiers(statements_a, statements_b,
        should_disambiguate_name=None):
    if should_disambiguate_name is None:
        def should_disambiguate_name(name):  # pylint:disable=function-redefined
            return True

    from pymbolic.imperative.analysis import get_all_used_identifiers

    id_a = get_all_used_identifiers(statements_a)
    id_b = get_all_used_identifiers(statements_b)

    from pytools import UniqueNameGenerator
    vng = UniqueNameGenerator(id_a | id_b)

    from pymbolic import var
    subst_b = {}
    for clash in id_a & id_b:
        if should_disambiguate_name(clash):
            unclash = vng(clash)
            subst_b[clash] = var(unclash)

    from pymbolic.mapper.substitutor import SubstitutionMapper, make_subst_func
    subst_map = SubstitutionMapper(make_subst_func(subst_b))

    statements_b = [
            stmt.map_expressions(subst_map) for stmt in statements_b]

    return statements_b, subst_b

# }}}


# {{{ disambiguate_and_fuse

def disambiguate_and_fuse(statements_a, statements_b,
        should_disambiguate_name=None):
    statements_b, subst_b = disambiguate_identifiers(
            statements_a, statements_b,
            should_disambiguate_name)

    fused, old_b_id_to_new_b_id = \
            fuse_statement_streams_with_unique_ids(
                    statements_a, statements_b)

    return fused, subst_b, old_b_id_to_new_b_id

# }}}


# vim: foldmethod=marker
